# Linux Application Checker
# Application Analysis Module (Analysis.pm)
#
# Copyright (C) 2007-2008 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

package Analysis;

use strict;
use File::Basename;
use File::Temp qw/ tempfile tempdir /;
use Common;

#-----------------------------------------------------------------------

# Global data structures

## Distributions
# $all_distrs->{$id} = {
#		ID, NAME, VER, FULLNAME, YEAR, VENDOR,
#		NO_GNU_HASH => 1, # if the .gnu.hash ELF section is not supported
#	};
our $all_distrs = {};

## Application Modules:
## (Fields marked with ~ are temporary or auxiliary.)
# $app_components->{$name} = {
#		NAME => $module, FILE, ABITAG,
#		NO_SHT_HASH, # if has no .hash ELF section
#		SONAME # for libraries included into the application
#		NEEDED_LIBS => [$soname],
#		NEEDED_INTS => {"$Iname\@$Iver" => 1},
#		$Did => { # Per distribution
#			NEEDED_LIBS_IMPL => [$soname], # list of implicitly needed libs, i.e. needed by needed libs.
#			LIBS->{$soname} => {
#				FOUND => (1|$distrlib), ~USED, ~STATUS
#				FILE, DISTR_COMPONENT, INTERFACES => $n,
#				VIA => $soname, # for libraries needed by a library from the NEEDED list
#			},
#			FOUND_INTS->{$fullname} => "$soname\|$ver_found",
#			MISSING_INTS->{$fullname} => "$soname", # $soname - guessed library for this interface
#			~INT_STATUS->{$fullname} => $status,
#			LIBS_OTHER_VERSION->{$soname_req} => {SONAME, FILE, DISTR_COMPONENT}, # not implemented yet
#		}
#		~LIB_STATUS->{$soname} => $status,
#		~LIBS_UNNEEDED->{$soname} => 1|0,
#		~LSB_STATUS => {LIBS->{$soname} => $status_obj, INTS->{$fullname} => $status_obj, },
#	};
our $app_components = {};

# Number of components skipped due to improper architecture
our $improper_arch = 0;

# Other information about libraries
our $approved_libs = {}; # $approved_libs->{$soname} = 1;

# Unneeded libraries -- used for distr status, etc.
our $libs_unneeded = {};

# Other information about interfaces
our $unknown_ints = {}; # $unknown_ints->{$Iname} = 1; # unknown
our $unmangled = {}; # $unmangled->{$Iname} = $unmangled_name;

## LSB Libraries:
# $lsb_libs->{$soname} -> {
#		SONAME, LSBNAME,
#		APPEAREDIN, MANDATORYSINCE, WITHDRAWNIN, DEPRECATEDSINCE, CANDIDATEFOR,
#		ATTR => {$LAtype => $LAvalue},
#	};
our $lsb_libs = {};

# $lsb_lib_status->{$soname} => {STATUS, STATUS_VER};
our $lsb_lib_status = {};

# $lib_attrs->{$soname}{$LAtype} => $LAvalue;
our $lib_attrs = {};

## LSB Interfaces:
# $lsb_ints->{$Iname}{$ver}{$lib} -> {
#		NAME, VER, LIBRARY,
#		APPEAREDIN, MANDATORYSINCE, WITHDRAWNIN, DEPRECATEDSINCE, CANDIDATEFOR,
#	};
our $lsb_ints = {};

# $lsb_int_status->{$Iname}{$soname} => {STATUS, STATUS_VER, ~REASON};
our $lsb_int_status = {};

# $int_attrs->{$Iname}{$lib}{$LAtype} => $LAvalue;
our $int_attrs = {};

# Auxiliary data:

## Libraries scanned from a current distribution (to obtain some info on unknown libraries)
# $loclibs->{$soname}{$Iname} = 1;
our $loclibs = {};

## Libraries provided by distributions:
# $distrlibs->{$distr}{$soname}[] -> {
#		POS, FILE, ABITAG, DISTR_COMPONENT,
#		IFACE_INTERVALS->{FIRST => [], SECOND => []}
#		~USED,
#	};
our $distrlibs = {};

# Library Knowledge Base
# $libKB->[$LKBid] = $text;
# $libKBDistr->{$soname}{$LKBid}{$group}{$dID} = 1;
our $libKB = [];
our $libKBDistr = {};

# Interface Knowledge Base
# $intKB->[$IKBid] = $text;
# $intKBDistr->{$Iid}{$IKBid}{$group}{$dID} = 1;
my $intKB = [];
my $intKBDistr = {};

my $save_format_ver = "3";

#-----------------------------------------------------------------------

my $crc32_poly = 0xEDB88320;

sub _crc32 {
	my $comp = shift;
	for (my $cnt = 0; $cnt < 8; $cnt++) { $comp = $comp & 1 ? $crc32_poly ^ ($comp >> 1) : $comp >> 1; }
	return $comp;
}

sub crc32 {
	my ($tcmp) = @_;
	my $crc = 0xFFFFFFFF;
	foreach (split(//,$tcmp)) { $crc = (($crc>>8) & 0x00FFFFFF) ^ _crc32(($crc ^ ord($_)) & 0xFF); }
	return $crc^0xFFFFFFFF;
}

sub load_distrs {
	$all_distrs = read_Distributions(@_);
}

# 
my %soname_to_libname_cache = ();

sub soname_to_libname {
	my ($soname) = @_;

	local $_;
	$_ = $soname_to_libname_cache{$soname} and return $_; # cache lookup
	
	return $soname_to_libname_cache{$soname} = soname_to_libname_sub($soname);
}

sub soname_to_libname_sub {
	my ($soname) = @_;

	my $lname = "";
	my $lversion = "";
	my $so_ver = "";
	
	$soname =~ s/.*\///; # remove the path

	$soname =~ s/\.so\.?(.*)//;
	$so_ver = $1;
	
	if( $soname =~ /^(ISO|GOST|UTF|TIS|ISIRI)/ ) {
		$lname = $soname;
	}
	elsif( $soname =~ /^(lib(?:g|stdc)\+\+-(?:\d-)?libc)/ ) {
		$lname = $1;
		
		#$soname =~ s/\.so\.?(.*)//;
		#$so_ver = $1;
		#$lname =~ s/\+\+-/"++-".$so_ver."-"/e;
	}
	else {
		$lname = $soname;
		
		if ( $lname =~ s/(?>-)(\d+.*)// ) {
			$lversion = $1;
		}
	}
	
	if ( $lname =~ /^lib(glib|gobject|gmodule|gthread|gdk_pixbuf|gdk_pixbuf_xlib|gdk-x11|gtk-x11)$/ ) {
		if ( $so_ver eq "2" || $lversion eq "2.0" ) {
			$lname .= '-2.0';
		}
	}
	elsif ( $lname =~ /^lib(atk|pango|pangocairo|pangoxft|pangoft2)$/ ) {
		if ( $lversion eq "1.0" ) {
			$lname .= '-1.0';
		}
	}
	elsif ( $lname =~ /^libgdk_pixbuf(_xlib)?$/ ) {
		if ( $lversion eq '2' ) {
			$lname .= '-2.0';
		}
	}
	elsif( $lname eq 'libstdc++' ) {
        $lname = 'libstdcxx';
    }
	elsif( $lname =~ /^ld-linux.*/ || $lname =~ /^ld64$/ ) {
        $lname = 'ld';
    }
	elsif ( $lversion =~ /^\d$/ ) {
		$lname .= '-'.$lversion;
	}
	
	return $lname;
}

# This function reads the Distributions file, which contains distro ID, name and other details.
sub read_Distributions {
	my ($filename) = @_;

	my $distrs = {};

	open FILE, $filename
		or do { print STDERR "Failed to open file '$filename'."; return undef; };

	while ( my $line = <FILE> ) {
		chomp $line;
		next if $line =~ /^#/; # skip comments

		# Format is simple: fields are separated by TAB.
		my ($id, $name, $ver, $year, $vendor, $gnu_hash) = split /\t/, $line;

		$distrs->{$id} = { ID => $id, NAME => $name, VER => $ver,
				FULLNAME => $name.' '.$ver, YEAR => $year, VENDOR => $vendor };
		$distrs->{$id}{NO_GNU_HASH} = 1 if defined $gnu_hash && !$gnu_hash;
	}

	return $distrs;
}
#-----------------------------------------------------------------------

sub save_lib_kb {
	my ($kb_file) = @_;

	# Lib Knowledge Base
	open FILE, ">".$kb_file
		or die "Failed to open file for writing: '$kb_file': $!";

	foreach (my $LKBid=0; $LKBid < @$libKB; $LKBid++ ) {
		print FILE "\n";
		print FILE $libKB->[$LKBid]."\n";

		foreach my $soname ( keys %$libKBDistr ) {
			next if !$libKBDistr->{$soname}{$LKBid};
			print FILE "LIB: $soname ";
			my @groups = ();
			foreach my $group ( keys %{$libKBDistr->{$soname}{$LKBid}} ) {
				push @groups, join ",", @{$libKBDistr->{$soname}{$LKBid}{$group}};
			}
			print FILE join ";", @groups;
			print FILE "\n";
		}
	}

	close FILE;
}

#-----------------------------------------------------------------------

# IntKB still not implemented

#-----------------------------------------------------------------------

sub read_LibAttrs {
	my ($filename) = @_;

	# Prepare the list of needed libraries
	my %req_sonames = ();
	foreach my $com ( values %$app_components ) {
		foreach my $soname ( @{$com->{NEEDED_LIBS}} ) {
			$req_sonames{$soname} = 1;
		}
	}
	
	open FILE, $filename
		or die "Failed to open file '$filename': $!";

	while( my $line = <FILE> ) {
		next if $line =~ /^#/;
		chomp $line;
		# FORMAT: LAsoname LAtype LAvalue

		my ($LAsoname, $LAtype, $LAvalue) = split /\t/, $line;

		next if !$req_sonames{$LAsoname};

		$LAvalue =~ s/\x81/\n/g; # \x81 is used instead of \n.

		if ( defined $lib_attrs->{$LAsoname}{$LAtype} ) {
			$lib_attrs->{$LAsoname}{$LAtype} .= " ".$LAvalue;
		} else {
			$lib_attrs->{$LAsoname}{$LAtype} = $LAvalue;
		}
	}
	close FILE;
}
#-----------------------------------------------------------------------

# This function reads a libKB file, which contains information
#  from the library Knowledge Base.
sub read_libKB {
	my ($filename) = @_;

	my $LKBid = 0;

	open FILE, $filename
		or die "Failed to open file '$filename': $!";

	my $mode = "text";
	my $text = "";
	while ( my $line = <FILE> ) {
		chomp $line;
		next if $line =~ /^#/; # skip comments

		# Format: a multiline text followed by one or more "LIB: " lines.
		# "LIB: " line specifies a library and few behaviour groups of distributions.
		if ( $mode eq "text" ) {
			if ( $line =~ /^LIB: / ) {
				$mode = "lib";
			}
		}
		elsif ( $mode eq "lib" ) {
			if ( $line !~ /^LIB: / ) {
				$mode = "text";
				++$LKBid;
				$text = "";
			}
		}

		if ( $mode eq "text" ) {
			next if ( $line eq "" && $text eq "" ); # Skip empty lines in the beginning of the message

			$text .= "\n" if $text ne "";
			$text .= $line;
		}
		elsif ( $mode eq "lib" ) {
			if ( !defined $libKB->[$LKBid] ) {
				$libKB->[$LKBid] = $text;
			}

			my ( $tmp, $soname, $distrs_grouped ) = split /\s+/, $line, 3;

			$distrs_grouped =~ s/\s+//g;

			# Distro groups are separated by semicolon ';'
			my @groups = split /;/, $distrs_grouped;
			foreach my $i ( 0 .. $#groups ) {
				# Distros in a group are separated by comma ','
				my @IDs = split /,/, $groups[$i];
				if ( @IDs ) {
					$libKBDistr->{$soname}{$LKBid}{$i+1} = [ @IDs ];
				}
			}
		}
	}
	close FILE;
}
#-----------------------------------------------------------------------

sub read_LSB_Libs {
	my ($filename) = @_;

	# Prepare the list of needed libraries
	my %req_sonames = ();
	foreach my $com ( values %$app_components ) {
		foreach my $soname ( @{$com->{NEEDED_LIBS}} ) {
			$req_sonames{$soname} = 1;
		}
	}

	# read the file
	open FILE, $filename
		or die "Failed to open file '$filename': $!";

	while( my $line = <FILE> ) {
		next if $line =~ /^#/;
		chomp $line;

		my @keys = qw/LSBNAME SONAME APPEAREDIN MANDATORYSINCE WITHDRAWNIN DEPRECATEDSINCE CANDIDATEFOR/;
		my @line = split /\t/, $line;
		shift @line; # shift LSBID

		# Filter needed libraries only
		my $soname = $line[1];
		next if !$req_sonames{$soname};

		my $lsb_lib = $lsb_libs->{$soname} ||= {};
		foreach ( my $i = 0; $i < @line; $i++ ) {
			$lsb_lib->{$keys[$i]} = $line[$i];
		}
	}
	close( FILE );
}
#-----------------------------------------------------------------------

# Read LSB interface list
sub read_LSB_Ints {
	my ($filename) = @_;

	# Prepare a list of needed interfaces
	my %req_interfaces = ();
	foreach my $com ( values %$app_components ) {
		if ( $com->{NEEDED_INTS} ) {
			foreach my $fullname ( keys %{$com->{NEEDED_INTS}} ) {
				my ($Iname, $ver) = split /\@/, $fullname; # Extract the interface name
				$req_interfaces{$Iname} = 1;
			}
		}
	}

	open FILE, $filename
		or die "Failed to open file '$filename'.";

	while( my $line = <FILE> ) {
		next if $line =~ /^#/;
		chomp $line;

		# First, split out the two first fields
		my ( $Iname, $ver, $other ) = split /\t/, $line, 3;

		# Check whether this interface is required by the Application
		next if !$req_interfaces{$Iname};

		# Split the rest
		my %d = ();
		@d{ qw/GENERIC LIBRARY APPEAREDIN MANDATORYSINCE WITHDRAWNIN DEPRECATEDSINCE CANDIDATEFOR/ }
			= split /\t/, $other;

		my $library = $d{LIBRARY} ||= "";

		my $lsb_int = $lsb_ints->{$Iname}{$ver}{$library} ||= {};

		if ( defined $lsb_int->{NAME} ) { # Already added

			# Entry with a specified architecture overrides an entry with a Generic arch
			next if ( $d{GENERIC} && !$lsb_int->{GENERIC} );
			if ( !$d{GENERIC} && $lsb_int->{GENERIC} ) {
				delete $lsb_int->{$_} for qw/GENERIC APPEAREDIN WITHDRAWNIN DEPRECATEDSINCE/;
			}

			# An interface can appear, be withdrawn and then appear again!
			# >:-F (e.g., clock_gettime, fstatfs64, inet_aton)
			# Merge the data
			for ( qw/APPEAREDIN WITHDRAWNIN DEPRECATEDSINCE/ ) {
				next if !defined $d{$_};
				if ( $lsb_int->{$_} ) {
					$lsb_int->{$_} .= ','.$d{$_}; # two or more values
				} else {
					$lsb_int->{$_} = $d{$_};
				}
			}

			# MANDATORYSINCE & CANDIDATEFOR fields should be the same by the construction of the file.
		} else {
			@$lsb_int{qw/NAME VER LIBRARY/} = ($Iname, $ver, $library); # always set
			for (qw/GENERIC APPEAREDIN MANDATORYSINCE WITHDRAWNIN DEPRECATEDSINCE CANDIDATEFOR/) {
				$lsb_int->{$_} = $d{$_} if defined $d{$_}; # set the optional fields
			}
		}
	}
	close( FILE );

	return undef;
}
#-----------------------------------------------------------------------

sub read_IntAttrs {
	my ($attrs_file) = @_;

	# Prepare a list of needed interfaces
	my %req_interfaces = ();
	foreach my $com ( values %$app_components ) {
		foreach my $Did ( keys %$all_distrs ) {
			foreach my $fullname ( keys %{$com->{$Did}{FOUND_INTS}} ) {
				my ($Iname, $ver) = split /\@/, $fullname; # Extract the interface name
				my $soname = $com->{$Did}{FOUND_INTS}{$fullname};
				$soname =~ s/\|.*//;
				
				my $lib = $com->{$Did}{LIBS}{$soname};
				if ( $lib && $lib->{VIA} ) { $soname = $lib->{VIA}; }
				
				$req_interfaces{$Iname}{soname_to_libname($soname)} = 1;
			}
			foreach my $fullname ( keys %{$com->{$Did}{MISSING_INTS}} ) {
				my ($Iname, $ver) = split /\@/, $fullname; # Extract the interface name
				my $soname = $com->{$Did}{MISSING_INTS}{$fullname};
				
				my $lib = $com->{$Did}{LIBS}{$soname};
				if ( $lib && $lib->{VIA} ) { $soname = $lib->{VIA}; }
				
				$req_interfaces{$Iname}{soname_to_libname($soname)} = 1;
			}
		}
	}
	
	open FILE, $attrs_file
		or die "Failed to open file '$attrs_file': $!";

	while( my $line = <FILE> ) {
		next if $line =~ /^#/;
		chomp $line;
		# FORMAT: IAiname IAlibrary IAtype IAvalue

		my ($IAiname, $IAlibrary, $IAtype, $IAvalue) = split /\t/, $line;

		next if !$req_interfaces{$IAiname};

		$IAvalue =~ s/\x81/\n/g; # \x81 is used instead of \n.

		foreach my $library ( keys %{$req_interfaces{$IAiname}} ) {
			next if $library && $IAlibrary && $library ne $IAlibrary;
				
			if ( defined $int_attrs->{$IAiname}{$library}{$IAtype} ) {
				$int_attrs->{$IAiname}{$library}{$IAtype} .= " ".$IAvalue;
			} else {
				$int_attrs->{$IAiname}{$library}{$IAtype} = $IAvalue;
			}
		}
	}
	close FILE;
}
#-----------------------------------------------------------------------

sub read_LibIndex {
	my ($filename, $filename2) = @_;

	# Read library index file
	open FILE, $filename
			or die "Failed to open file '$filename'.";

	while ( my $line = <FILE> ) {
		next if $line =~ /^#/;
		chomp $line;

		my ($soname, $pos, $RLid, $distr, $file, $abitag) = split /,/, $line;

		$abitag = "" if not defined $abitag;

		my $distrlib = { ID => $RLid, POS => $pos, FILE => $file, ABITAG => $abitag };

		push @{$distrlibs->{$distr}{$soname}}, $distrlib;
	}

	close FILE;

	# Read unapproved libraries
	open FILE, $filename2
			or die "Failed to open file '$filename2'.";

	while ( my $line = <FILE> ) {
		next if $line =~ /^#/;
		chomp $line;

		my ($soname, $RLid, $distr, $file, $abitag) = split /,/, $line; # without POS

		$abitag = "" if not defined $abitag;

		my $distrlib = { ID => $RLid, FILE => $file, ABITAG => $abitag }; # without POS

		push @{$distrlibs->{$distr}{$soname}}, $distrlib;
	}

	close FILE;
}

# Load library dependencies
sub read_RLibDeps {
	my ($filename) = @_;
	
	open FILE, $filename
			or die "Failed to open file '$filename'.";

	my $deps = [];
	
	use constant ONE_BYTE_LIM => 30;
	
	my $n = 0;
	while ( my $line = <FILE> ) {
		$n++;
		next if $line =~ /^$/;
		chomp $line;
		
		my @line = split //, $line;
		while ( @line ) {
			my $a = shift @line;
			$a = ord($a) - 32;
			if ( $a >= ONE_BYTE_LIM ) {
				$a -= ONE_BYTE_LIM;
			} else {
				my $b = shift @line;
				( defined $b ) or die;
				$b = ord($b) - 32;
				$a = $a * 95 + $b;
			}
			push @{$deps->[$n]}, $a;
		}
	}

	close FILE;
	
	return $deps;
}

# This function reads the Sonames file, which contains soname IDs used in library dependencies.
sub read_Sonames {
	my ($filename) = @_;

	my $sonames = [];

	open FILE, $filename
		or do { print STDERR "Failed to open file '$filename'."; return undef; };

	while ( my $line = <FILE> ) {
		chomp $line;
		next if $line =~ /^#/; # skip comments

		# Format is simple: fields are separated by TAB.
		my ($soname, $id) = split /\t/, $line;

		$sonames->[$id] = $soname;
	}

	return $sonames;
}

# Load index for the Interfaces_<ARCH> file
sub read_IIndex {
	my ($filename) = @_;

	my @int_index = ();

	open FILE, $filename
		or die "Failed to open file '$filename'.";

	# Format: line number is an index beginning from 1000;
	# A number in each line is a position in Interfaces file.

	my $idx = 0;
	while ( my $line = <FILE> ) {
		chomp $line;

		$int_index[$idx] = $line;
		$idx++;
	}
	close FILE;

	return \@int_index;
}

# Pick up interface IDs from the IIndex file.
sub find_interfaces {
	my ($filename, $int_index) = @_;

	# Prepare a list of needed interfaces
	my %req_ints = ();
	foreach my $com ( values %$app_components ) {
		next if !$com->{NEEDED_INTS};
		foreach my $fullname ( keys %{$com->{NEEDED_INTS}} ) {
			my ($Iname, $ver) = split /\@+/, $fullname;
			$ver ||= "";
			
			my $key = substr( crc32($Iname), 0, 4) - 1000;

			$req_ints{$key}{$Iname}{$ver} = 1;
		}
	}

	my $interfaces_fh;
	open $interfaces_fh, $filename
		or die "Failed to open file '$filename'.";

	print "!find_interfaces: ".scalar(keys %req_ints)."\n";

	my $riids = {}; # $riids->{$Iname}{$ver} = [ IDs ];

	my $last_pos = 0;
	# Interfaces to be searched are enumerated in the same order as in the file.
	# So we could read the file incrementally. This should enhance performance.
	foreach my $key ( sort {$a <=> $b} keys %req_ints ) {
	foreach my $Iname ( sort keys %{$req_ints{$key}} ) {
	# Read all versions of this interface
	#foreach my $ver ( sort keys %{$req_ints{$key}{$Iname}} ) {
		my $pos = $int_index->[$key];
		my $max_pos = $int_index->[$key+1];

		if ( $pos < $last_pos ) {
			# Search from the position where stopped at the previous iteration.
			$pos = $last_pos;
		}

		my ($found_pos, %IDs) = find_interface( $interfaces_fh, $Iname, $pos, $max_pos );

		$last_pos = $found_pos if %IDs;

		if ( %IDs ) { 
			foreach my $found_ver ( keys %IDs ) {
				$riids->{$Iname}{$found_ver} = [sort {$a<=>$b} @{$IDs{$found_ver}}];
			}
		}
	}}

	close $interfaces_fh;

	# Unknown interfaces
	foreach my $key ( sort {$a <=> $b} keys %req_ints ) {
	foreach my $Iname ( sort keys %{$req_ints{$key}} ) {
		if ( !$riids->{$Iname} ) {
			$unknown_ints->{$Iname} = 1; # TODO: test it
		}
		# TODO: other versions found
	}}

	return $riids;
}

# This function searches fro interface ID in the Interfaces file.
# Since the file is very large (over 30Mb), the index (IIndex file) is used.
# The Interfaces file consists of groups of interfaces with same crc sum of their names.
# Interfaces in a group sorted alphabetically by name.

# This function returns position of the last found element and an array
# of found interfaces (there could be several records for one interface name).
# The position of the last found element is used as a beginning position
# for the further search.
sub find_interface {
	my ($fh, $Iname, $pos, $max_pos) = @_;

	die if !defined $Iname;

	seek $fh, $pos, 0;

	my %result = ();

	my $line_len;
	my $prev_name = "";
	while ( my $line = <$fh> ) {
		$line_len = length($line);

		last if $max_pos && $pos >= $max_pos;

		chomp $line;

		if ( $line =~ /^([^,;]*)(,([^,;]+))?;(\d+)$/ ) {
			my $name1 = $1;
			my $ver1 = $3 || "";
			my $Iid = $4;

			if ( $name1 eq "" ) {
				$name1 = $prev_name;
			}
			$prev_name = $name1;

			next if $name1 lt $Iname; # continue searching

			last if $name1 gt $Iname; # no more

			# Gotcha it if reached this line.
			push @{$result{$ver1}}, $Iid;
		}
		else {
			die "Interfaces: wrong line '$line'";
		}
	}
	continue {
		$pos += $line_len;
	}

	return ($pos, %result);
}
#-----------------------------------------------------------------------

sub read_upload_data {
	my ($filename, $architecture) = @_;

	my %provided_libs = ();
	my %provided_ints = ();

	open FILE, $filename
		or die "Failed to open file for reading '$filename': $!";

	my $stat = {};

	my $line;
	my $mode = '';
	my $com = undef; # An application module being processed
	while( $line = <FILE> ) {
		chomp $line;
		
		if ( $mode eq '' ) { # top-level mode
			if( $line =~ /^\!Command/ ) {
				# The string should contain name and path
				if ( $line =~ /^\!Command '([^']+)' '([^']*)'/ ) {
					my ($real_file, $com_name ) = split /\|/, $2, 2;

					$com = { NAME => $com_name,	FILE => $real_file,
					         NEEDED_LIBS => [], NEEDED_INTS => {}
						};
					
					$app_components->{$com_name} = $com;
					
					$mode = 'com';
				}
				else { warn "Incorrect command record at line $.:\n\t'$_'"; }
			}
			elsif ( $line =~ /^\!Library/ ) {
				# The string should contain name, version, runname, path, ABI tag (optional) and so-name (optional) of the library
				if ( $line =~ /^\!Library '([^']+)' '([^']*)' '([^']+)' '([^']+)' '([^']*)' '([^']*)'/ ) {
					my $runname = $3;
					my $origin_name = "$4/$runname";
					my $app_abitag = $5;
					my $soname = $6;
					
					# Remember this lib as provided by the app
					$provided_libs{$runname} = 1;
					if( $runname =~ /(.+\.so)\./ ) {
						$provided_libs{$1} = 1;
					}
					if( $soname ) {
						$provided_libs{$soname} = 1;
					}

					my ($real_file, $com_name ) = split /\|/, $origin_name, 2;
					
					$com = { NAME => $com_name, FILE => $real_file,
					         NEEDED_LIBS => [], NEEDED_INTS => {}
						};
					$com->{ABITAG} = $app_abitag if $app_abitag;
					$com->{SONAME} = $soname;
					
					$app_components->{$com_name} = $com;
					
					$mode = 'com';
				}
				else { warn "Incorrect library record at line $.:\n\t'$_'"; }
			}
		}
		elsif ( $mode eq 'com' ) {
			if ( $line =~ /^ELF Header/ ) {
				$mode = 'elfheader';
			}
			elsif ( $line =~ /^Section Headers:/ ) {
				$mode = 'secheaders';
			}
			elsif ( $line =~ /^Dynamic section/ ) {
				$mode = 'dynsec';
			}
			elsif ( $line =~ /^\S/ and $line =~ /'\.dynsym'/ ) {
				$mode = 'dynsym';
			}
			# symtab is ignored
			elsif ( $line =~ /^\!/ ) {
				$mode = '';
				redo;
			}
		}
		elsif( $mode eq 'elfheader' ) {
			if ( $line =~ /^\s+Class\:\s*ELF(\d+)/ ) {
				$com->{'elf_arch_bit'} = $1; # do we deal with 32 or 64 bit ELF file
				next;
			}
			elsif ( $line =~ /^\s+Machine\:\s*(.+)/ ) {
				my $elf_arch = $1;
				
				if ( !is_elf_arch_correct($architecture, $elf_arch, $com->{'elf_arch_bit'}) ) {
					$improper_arch++;
					delete $app_components->{$com->{NAME}};
					$com = undef;
					$mode = '';
				}
			}
			elsif ( $line =~ /^\s*$/ ) { # empty line
				delete $com->{'elf_arch_bit'};
				$mode = 'com';
			}
		}
		elsif ( $mode eq 'secheaders' ) {
			if ( $line =~ /^\s+\[[ \d]*\]\s+\.hash\s/ ) {
				$com->{'sht_hash'} = 1; # whether the component has .hash section
				next;
			}
			elsif ( $line =~ /^\s*$/ ) { # empty line
				if ( !$com->{'sht_hash'} ) {
					$com->{NO_SHT_HASH} = 1;
				} else {
					delete $com->{'sht_hash'};
				}
				$mode = 'com';
			}
		}
		elsif( $mode eq 'dynsec' ) {
			if( $line =~ /NEEDED\)\s+Shared\s+library\:\s+\[([^\]]+)\]/ ) {
				my $libname = $1;
				
				if ( $libname =~ m{/([^/]+)/?$} ) { # TODO: this is not very good
					$libname = $1;
				}
				
				push @{$com->{NEEDED_LIBS}}, $libname;
			}
			elsif ( $line =~ /^\s*$/ ) { # empty line
				delete $com->{'elf_arch_bit'};
				$mode = 'com';
			}
		}
		elsif( $mode eq 'dynsym' ) {
			# (the line of 'readelf' output corresponding to the interface)
			if ( $line =~ /\s*\d+:\s+\w*\s+\w+\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s(\S+)/ ) {
				my $type = $1;	# FUNC, OBJECT etc.
				my $bind = $2;	# GLOBAL, WEAK or LOCAL
				my $vis = $3;	# Visibility
				my $Ndx = $4;	#
				my $fullname = $5;	# name, maybe with version

				# Filter interfaces by their binding, type and visibility
				next if $bind ne 'WEAK' and $bind ne 'GLOBAL';
				next if $type ne 'FUNC' and $type ne 'OBJECT' and $type ne 'COMMON';
				next if $vis  ne 'DEFAULT';

				$fullname =~ s/\@\@/\@/;
				my ($Iname, $Iver) = split /\@/, $fullname;
				$Iver = "" if !$Iver;

				if( $Ndx eq "UND" ) {
					# Required interface
					$com->{NEEDED_INTS}->{$fullname} = 1;
				}
				else {
					# Internal interface
					$provided_ints{$fullname} = 1;
					$provided_ints{$Iname} = 2 if !$provided_ints{$Iname};
				}
			}
			elsif ( $line =~ /^\s*$/ || $line =~ /^\S/ ) { # Empty line, or without leading space
				$mode = 'com';
				redo;
			}
		}
	}

	# Delete provided libs and shared ints
	foreach my $app_module ( values %$app_components ) {
		# Delete provided libs
		@{$app_module->{NEEDED_LIBS}} = grep {!$provided_libs{$_}} @{$app_module->{NEEDED_LIBS}};

		# Delete shared ints
		if ( defined $app_module->{NEEDED_INTS} ) {
			foreach my $fullname ( keys %{$app_module->{NEEDED_INTS}} ) {
				my $Iname = $fullname;
				$Iname =~ s/\@.*//; # remove version
				if (   $provided_ints{$fullname} # fullname matches
					|| $provided_ints{$Iname} && $provided_ints{$Iname} == 1 ) # interface without version is provided
				{
					delete $app_module->{NEEDED_INTS}->{$fullname};
				}
			}
		}
	}
	close FILE;
}

sub find_libifaces {
	my ($filename) = @_;

	# Prepare a list of offsets where the LibIfaces file should be read.
	# Offsets was previously read from the LibIndex file.
	my %by_pos = ();
	foreach my $Did ( keys %$distrlibs ) {
		foreach my $soname ( keys %{$distrlibs->{$Did}} ) {
			foreach my $distrlib ( @{$distrlibs->{$Did}{$soname}} ) {
				next if not $distrlib->{USED};
				next if not defined $distrlib->{POS};
				$by_pos{ $distrlib->{POS} } = $distrlib;
			}
		}
	}

	my $libifaces_fh;
	open $libifaces_fh, $filename
		or die "Failed to open file '$filename'.";

	foreach my $pos ( sort keys %by_pos ) {
		# Sequentially read the LibIfaces file
		read_LibIfaces( $libifaces_fh, $pos, $by_pos{$pos} );
	}

	close $libifaces_fh;
}

sub read_LibIfaces {
	my ($fh, $pos, $distrlib_obj) = @_;

	seek $fh, $pos, 0
		or die "read_LibIfaces: seek to '$pos' failed";

	my $a = [];
	my $b = [];

	# Read Interface IDs
	# Each library is on one line.
	my $line = <$fh>;
	chomp $line;
	last if $line eq ""; # no interfaces

	# Interface ID intervals
	# IDs are stored in a simple compressed format:
	# The first ID is written as is.
	# Each next number is in format of "+x" or "-y";
	# These are included and excluded intervals:
	# +x means that the next x IDs are included.
	# -y means that the next y IDs are excluded.
	# 1 in +1 and -1 is omitted.
	# So the sequence (55,56,57,70,71,90,100) is encoded as 55+3-12+2-18+-9+
	# Real IDs have up to 7 digits so the compression is really useful.
	my $first;
	my $last;
	if ( $line =~ s/^(\d+)// ) {
		$first = $last = int($1);
	} else {
		die "No base in line $line";
	}

	while ( $line =~ s/^([\+\-])(\d*)// ) {
		my $pm = $1;
		my $d = $2;
		$d = 1 if !$d;

		if ( $pm eq '+' ) {
			$last = $first + $d - 1;
			push @$a, $first;
			push @$b, $last;
		}
		elsif ( $pm eq '-' ) {
			$last = $first = $last + $d + 1;
		}
		else {
			die "Syntax error"; # should not happen
		}
	}

	$distrlib_obj->{IFACE_INTERVALS} = {FIRST=>$a, SECOND=>$b};

	return undef;
}

sub link_libs {
	my ($deps, $sonames) = @_;
	
	my $cache = {};

	foreach my $com ( values %$app_components ) {
		my $com_abitag = ($com->{ABITAG} or "");
		
		foreach my $Did ( keys %$all_distrs ) {
			my $cache_d = undef;
			if ( defined $cache->{$com_abitag, $Did} ) {
				$cache_d = $cache->{$com_abitag, $Did};
			}
			
			$com->{$Did}{NEEDED_LIBS_IMPL} = [];
			my $needed_libs_impl = {};
			
			my $distrlibs_d = $distrlibs->{$Did};
			next if !$distrlibs_d;
			
			# @to_search is initialized again for each distribution
			#  because it is changed during the cycle iteration
			my @to_search = @{$com->{NEEDED_LIBS}};
			push @to_search, $com->{SONAME} if $com->{SONAME}; # Want to know if this component is present in distributions
			
			my $i = 0;
			while ( $i < @to_search ) { # @to_search may be changed in the cycle
				my $soname = $to_search[$i];
			
				my $distr_lib;
				
				# $cache_d->{$soname} may be undefined but exist this means it wasn't found.
				if ( $cache_d && exists $cache_d->{$soname} ) { # 'exists' is important! ^^^
					$distr_lib = $cache_d->{$soname};
				}
				else {
					my $libs = $distrlibs_d->{$soname};
					
					next if !$libs || !@$libs;
					
					$distr_lib = linker_heuristic($libs, $com_abitag);
				}
				
				if ( $distr_lib ) {
					$distr_lib->{USED} = 1;
					
					unless ( $com->{SONAME} && $soname eq $com->{SONAME} ) {
						my $RLid = $distr_lib->{ID};
						if ( $deps->[$RLid] ) { # Add libraries needed by this library
							foreach my $sid ( @{$deps->[$RLid]} ) {
								my $dep_soname = $sonames->[$sid] or die;
								
								my $already_added = 0;
								foreach my $s ( @to_search ) {
									if ( $s eq $dep_soname ) {
										$already_added = 1;
										last;
									}
								}
								if ( !$already_added ) {
									push @to_search, $dep_soname;
									
									if ( !defined $needed_libs_impl->{$dep_soname} ) {
										push @{$com->{$Did}{NEEDED_LIBS_IMPL}}, $dep_soname;
										my $via = $needed_libs_impl->{$soname}; # Keep the top-level library as a parent for all secondary libraries
										$needed_libs_impl->{$dep_soname} = $via || $soname;
									}
								}
							}
						}
					}
					
					my $lib = $com->{$Did}{LIBS}{$soname} ||= {};
					$lib->{FOUND} = $distr_lib;
					$lib->{FILE} = $distr_lib->{FILE};
					$lib->{DISTR_COMPONENT} = $distr_lib->{DISTR_COMPONENT};
					if ( $needed_libs_impl->{$soname} ) {
						$lib->{VIA} = $needed_libs_impl->{$soname};
					}
					
					$approved_libs->{$soname} = 1 if $distr_lib->{POS};
				}

				# Remember the $distr_lib (or undef if wasn't found).
				$cache_d->{$soname} = $distr_lib;
			}
			continue { $i++; }
		}
	}
}

sub parse_abitag( $ ) {
	my ($abitag) = @_;

	if ( !defined $abitag || $abitag eq "" ) {
		return (0, 0, 0);
	}

	if ( $abitag =~ /^Linux\s+(\d+)\.(\d+)\.(\d+)\s*$/ ) {
		return ($1, $2, $3);
	}
	else {
		print STDERR "Warning! Can't parse ABI-tag '$abitag'\n";
		return (0, 0, 0);
	}
}

sub abitag_ge ( $ $ ) {
	my ($a, $b) = @_;

	if (   $a->[0] >  $b->[0]
		|| $a->[0] == $b->[0] && $a->[1] >  $b->[1]
		|| $a->[0] == $b->[0] && $a->[1] == $b->[1] && $a->[2] >= $b->[2] )
	{
		return 1;
	}

	return 0;
}

sub linker_heuristic {
	my ($libs, $app_abitag) = @_;

	if ( scalar @$libs == 0 ) { return ''; }
	if ( scalar @$libs == 1 ) { return $libs->[0]; }

	my @libs = @$libs; # Make a working copy
	my @libs_copy; # Temporary var

	# ( Explanation on ABI tags by Ulrich Drepper:
	#   http://people.redhat.com/drepper/assumekernel.html )

	if ( $app_abitag ) {
		my @app_abitag = parse_abitag($app_abitag);

		@libs_copy = ();

		my $select_min = 1;

		# Filter by abitag
		foreach my $lib ( @libs ) {
			if ( $lib->{ABITAG} ) {
				my @lib_abitag = parse_abitag( $lib->{ABITAG} );

				next if not abitag_ge( \@lib_abitag, \@app_abitag );
			} else {
				$select_min = 0;
			}

			push @libs_copy, $lib;
		}

		if ( scalar @libs_copy == 1 ) { return $libs->[0]; }

		if ( @libs_copy ) {
			@libs = @libs_copy;
		} else {
			$select_min = 0;
		}

		if ( $select_min ) {
			# Select min of filtered abitags
			my @min_abitag = (99,99,99);
			my $min_abitag = "";
			foreach my $lib ( @libs ) {
				my @lib_abitag = parse_abitag( $lib->{ABITAG} );

				if ( not abitag_ge( \@lib_abitag, \@min_abitag )  ) {
					@min_abitag = @lib_abitag;
					$min_abitag = $lib->{ABITAG};
				}
			}
			@libs = grep {$_->{ABITAG} eq $min_abitag} @libs;

			if ( scalar @libs == 1 ) { return $libs[0]; }
		}
	}

	# /lib and /usr/lib
	@libs_copy = grep {$_->{FILE} =~ m{^/lib(/|32/|64/)} } @libs;
	if ( @libs_copy ) {
		@libs = @libs_copy; # select from this new subset
	}
	if ( scalar @libs == 1 ) { return $libs[0]; }
	
	@libs_copy = grep {$_->{FILE} =~ m{^/usr/lib(/|32/|64/)} } @libs;
	if ( @libs_copy ) {
		@libs = @libs_copy;
	}
	if ( scalar @libs == 1 ) { return $libs[0]; }

	# Pick any one (the shortest path, for example)
	@libs = reverse sort {
			my ($A, $B) = ($a->{FILE}, $b->{FILE});
			($A =~ s!/!!g) <=> ($B =~ s!/!!g) # compare number of "/" in elements
			or $a->{FILE} cmp $b->{FILE};
		} @libs;
	return $libs[0];
}
#-----------------------------------------------------------------------

# Find which library is an interface from for each required interface.
sub link_ints {
	my ($riids) = @_;

	my $tmp_all_ints = {};
	foreach my $com (values %$app_components ) {
		foreach my $fullname ( keys %{$com->{NEEDED_INTS}} ) {
			my ($Iname, $ver) = split /\@/, $fullname, 2;
			$ver ||= "";
			
			my $int_riids = undef;
			
			# $int_riids should be sorted!
			
			if ( $riids->{$Iname} ) {
				if ( $ver ) {
					$int_riids = $riids->{$Iname}{$ver}; # Was sorted in find_interfaces();
				} else {
					# Any version will do
					$int_riids = [ sort {$a<=>$b} map {@$_} values %{$riids->{$Iname}} ]; # re-sort
				}
			}
			
			# Even if the $int_riids is undefined (unknown interface name):
			foreach my $Did ( keys %$all_distrs ) {
				my $com_distr = $com->{$Did};
				
				# Search for the interface in a given array of libraries
				local *search = sub {
					my ($needed_libs) = @_;
					
					foreach my $soname ( @$needed_libs ) {
						my $distrlib = $com_distr->{LIBS}{$soname}{FOUND};
						
						if ( $int_riids && $distrlib && $distrlib->{IFACE_INTERVALS} ) {
							# Search in the $distrlib

							# Interfaces of the current library grouped by intervals.
							my $first = $distrlib->{IFACE_INTERVALS}{FIRST};
							my $second = $distrlib->{IFACE_INTERVALS}{SECOND};

							# Look if one of the interface's IDs matches any of the intervals.
							my $riidx = 0;
							my $riid = $int_riids->[$riidx];
							for ( my $i=0; $i < @$first; $i++ ) {
								if ( $riid < $first->[$i] ) {
									$riidx++;
									last if $riidx > $#$int_riids;
									$riid = $int_riids->[$riidx];
									redo;
								}
								elsif ( $riid <= $second->[$i] ) {
									# gotcha!
									return $soname;
								}
							}
						}
					} # end of  foreach $soname ( ... )
					return undef; # Failed to find anything
				}; # end of  local sub search()

				# First, search for the interface in NEEDED_LIBS
				my $soname = search( $com->{NEEDED_LIBS} );
				$soname ||= search( $com_distr->{NEEDED_LIBS_IMPL} );
				
				if ( $soname ) { # found
					# TODO: keep found versions for interfaces with no version specified.
					$com_distr->{FOUND_INTS}{$fullname} = $soname; # TODO: |$found_ver ?
				} else { # not found
					# Let's look in the local libraries
					foreach my $lib ( @{$com->{NEEDED_LIBS}} ) {
						if ( $loclibs->{$lib}{$fullname} ) {
							$soname = $lib;
							last;
						}
					}
					if ( $soname ) {
						$com_distr->{MISSING_INTS}{$fullname} = $soname;
					}
				}
				
				if ( $soname ) {
					my $lib = $com_distr->{LIBS}{$soname};
					if ( $lib ) { $lib->{USED} = 1; } # Mark the lib as used
					if ( $lib && $lib->{VIA} ) { $soname = $lib->{VIA}; }
					$tmp_all_ints->{$Iname}{$ver}{$soname}++;
				}
			}
		}
	}
	
	# Guess library for unlinked interfaces by other components/distros.
	foreach my $com (values %$app_components ) {
		foreach my $fullname ( keys %{$com->{NEEDED_INTS}} ) {
			my ($Iname, $ver) = split /\@/, $fullname, 2;
			$ver ||= "";
			foreach my $Did ( keys %$all_distrs ) {
				if ( !$com->{$Did}{FOUND_INTS}{$fullname}
					&& !$com->{$Did}{MISSING_INTS}{$fullname} )
				{ # non-linked
					if ( $tmp_all_ints->{$Iname} ) {
						my $soname;
						# TODO: prefer LSB-compatible version and library
						if ( !$tmp_all_ints->{$Iname}{$ver} ) { # another version
							# Dirty hack: set $ver to another one
							my $h = $tmp_all_ints->{$Iname};
							$ver = (sort {$h->{$a} cmp $h->{$b}} keys %$h)[-1]; # max version
						}
						if ( $tmp_all_ints->{$Iname}{$ver} ) { # linked 
							my $h = $tmp_all_ints->{$Iname}{$ver};
							my $max = 0;
							my $best_soname = undef;
							for my $tmp_soname ( @{$com->{NEEDED_LIBS}} ) {
								next if !defined $h->{$tmp_soname};
								if ( $h->{$tmp_soname} > $max ) { # choose max used soname
									$max = $h->{$tmp_soname};
									$best_soname = $tmp_soname;
								}
							}
							$soname = $best_soname;
						}
						
						if ( $soname ) {
							$com->{$Did}{MISSING_INTS}{$fullname} = $soname;
						}
					}
				}
			}
		}
	}
}
#-----------------------------------------------------------------------

# Unmangle interface names
sub unmangle {
	# Make a list of mangled interface names,
	# then write it into a temporary file, then feed this file to the `c++filt`
	# and read the unmangled names from its output.

	my %int_names = ();
	foreach my $com (values %$app_components ) {
		foreach my $fullname ( keys %{$com->{NEEDED_INTS}} ) {
			my $Iname = $fullname;
			$Iname =~ s/\@.*//;
			$int_names{$Iname} = 1;
		}
	}

	# Don't sort names, just remember the order
	my $unmangle_order = [keys %int_names];

	# Create a temporary file
	my ($tempfh, $tempfilename) = tempfile("/tmp/appchk_unmangle_XXXXXX", UNLINK => 1);
	foreach ( @$unmangle_order ) {
		print $tempfh "$_\n";
	}
	close $tempfh;

	# Call c++filt
	$SIG{PIPE} = 'IGNORE'; # ignore this signal while reading from a pipe
	open CPPFILT, "cat '$tempfilename' | c++filt |";

	# Read the result from stdout
	foreach my $Iname ( @$unmangle_order ) {
		my $unmangled_name = <CPPFILT>;
		if ( !$unmangled_name ) {
			die "c++filt stopped when unmangled name of '$Iname' was expected." ;
		}
		chomp $unmangled_name;

		# Don't keep duplicates of initially non-mangled names:
		next if $unmangled_name eq $Iname;

		$unmangled->{$Iname} = $unmangled_name;
	}
	close CPPFILT;
	$SIG{PIPE} = 'DEFAULT';
}

sub get_ldconfig {
	my ($arch) = @_;

	my $grep = "";
	my $neg = 0;

	if( $arch =~ /x86-64/i ) {
		$grep = "x86-64";
	}
	elsif( $arch =~ /IA32/i or $arch =~ /x86/i ) {
		# x86 apps will run on x86-64, but they will use 32bit libs
		$grep = "x86-64";
		$neg = 1;
	}
	elsif( $arch =~ /IA64/i ) {
		$grep = "IA-64";
	}
	elsif( $arch =~ /S390X/i ) {
		$grep = "64bit";
	}
	elsif( $arch =~ /PPC64/i ) {
		$grep = "64bit";
	}
	elsif( $arch =~ /S390/i ) {
		$grep = "64bit";
		$neg = 1;
	}

	$SIG{PIPE} = 'IGNORE';
	open PIPE, "/sbin/ldconfig -p |";

	my $ld_config = {};

	while ( <PIPE> ) {
		chomp;

		my $test = /\b$grep\b/;
		my $wrong = 0;
		if ( $neg && $test or !$neg && !$test) {
			# Wrong bit number
			$wrong = 1;
		}

		if ( /\s+(\S+)\s[^=]*=>\s*(\S+)/ ) {
			my $libname = $1;
			my $realname = $2;

			if ( !$wrong || !$ld_config->{$libname} ) {

				$ld_config->{$libname} = $realname;

				# Also save short soname. It's better to find a wrong version
				# of the needed library than find nothing.
				$libname =~ s/\.so.*/.so/;
				if ( !$ld_config->{$libname} ) {
					$ld_config->{$libname} = $realname;
				}
			}
		}
	}
	close PIPE;
	$SIG{PIPE} = 'DEFAULT';

	return $ld_config;
}
#-----------------------------------------------------------------------

sub get_local_libs {
	my ($architecture) = @_;

	my %to_scan = ();

	foreach my $com (values %$app_components ) {
		# Scan even known libs, because an interface could be from a library
		# with an approved soname, but not in any library in the database.
		# Thus we will be able to identify which library is this interface from.
		foreach my $soname ( @{$com->{NEEDED_LIBS}} ) {
			$to_scan{$soname} = 1;
		}
	}

	return if !%to_scan;

	my $ld_config = get_ldconfig($architecture);

	foreach my $soname ( keys %to_scan ) {
		#print STDERR "Scanning local lib: $soname\n";
		scan_local_lib($soname, $ld_config);
	}
}
#-----------------------------------------------------------------------

# Extract interfaces from a local library
sub scan_local_lib {
	my ($soname, $ld_config) = @_;

	my $libname = $soname;
	my $realname = $ld_config->{$libname};
	if ( !$realname ) {
		# Try soname without version even if it's not an exact match
		$libname =~ s/(?<=\.so).*//;
		$realname = $ld_config->{$libname};
	}
	return if !$realname;

	# -WhlSsdA
	open( LIB_DATA, "LC_ALL=C readelf -Ws $realname | " );

	my $dynsym = 0;

	while( <LIB_DATA> ) {
		# We actually need only '.dynsym' section at the moment
		if( /^Symbol\s*table\s*'\.dynsym'/i ) {
			$dynsym = 1;
			next;
		}

		if( !$dynsym ) {
			next;
		}

		# Non-space in the beginning of the line; '.dynsym' section is finished
		if( /^\S/ ) {
			$dynsym = 0;
			last;
		}

		chomp;

		if( /\s*\d+:\s+\w*\s+\w+\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s((\w|@|\.)+)/ ) {
			# (the line of 'readelf' output corresponding to the interface)
			my $type=$1;	 # FUNC, OBJECT etc.
			my $bind=$2;	 # GLOBAL, WEAK or LOCAL
			my $vis=$3;	  # Visibility
			my $Ndx=$4;	  # UND for external interfaces
			my $fullname=$5; # name, maybe with version

			# Filter interfaces by their binding, type and visibility
			if( ($bind ne 'WEAK') and ($bind ne 'GLOBAL') ) {
				next;
			}
			if( ($type ne 'FUNC') and ($type ne 'OBJECT') and ($type ne 'COMMON') ) {
				next;
			}
			if( ($vis ne 'DEFAULT') or ($Ndx eq "UND") ) {
				next;
			}

			$fullname =~ s/\@\@/\@/;

			$loclibs->{$soname}{$fullname} = 1;
			my $Iname = $fullname;
			$Iname =~ s/\@.*//;
			$loclibs->{$soname}{$Iname} = 1;
		}
	}
	close( LIB_DATA );
}
#-----------------------------------------------------------------------

sub is_elf_arch_correct($$$ ) {
	my($arch, $elf_arch, $elf_arch_bit) = ($_[0],$_[1],$_[2]);

	if( $arch =~ /x86-64/i ) {
		if( $elf_arch !~ /86-64/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /IA64/i ) {
		if( $elf_arch !~ /IA-64/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /S390X/i ) {
		if( ($elf_arch !~ /390/) or ($elf_arch_bit ne "64") ) {
			return 0;
		}
	}
	elsif( $arch =~ /PPC64/i ) {
		if( $elf_arch !~ /PowerPC 64-bit/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /PPC32/i ) {
		if( ($elf_arch !~ /PowerPC/) or ($elf_arch =~ /PowerPC 64-bit/) or ($elf_arch =~ /PowerPC64/) ) {
			return 0;
		}
	}
	elsif( $arch =~ /x86/i ) {
		if( $elf_arch !~ /Intel 80386/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /S390/i ) {
		if( ($elf_arch !~ /390/) or ($elf_arch_bit ne "32") ) {
			return 0;
		}
	}
	else {
		print STDERR "is_elf_arch_correct(): Unknown arch: $arch\n";
	}

	return 1;
}
#-----------------------------------------------------------------------

sub save_main {
	my ($filename) = @_;

	open FILE, ">$filename"
		or do { print STDERR "Failed to open file for writing: '$filename': $!"; return undef; };

	print FILE "# Analysis results\n";

	print FILE "\nFORMAT: ".$save_format_ver."\n";
	
	foreach my $name ( sort keys %$app_components ) {
		# save the app component
		my $com = $app_components->{$name};
		print FILE "\n*APP_COMPONENT\n";
		for my $key ( qw/NAME FILE ABITAG NO_SHT_HASH SONAME/ ) {
			print FILE "$key: ".$com->{$key}."\n" if $com->{$key};
		}

		# NEEDED libraries
		print FILE "\n**NEEDED_LIBS\n";
		foreach my $soname ( @{$com->{NEEDED_LIBS}} ) {
			print FILE $soname."\n";
		}

		# NEEDED_LIBS_IMPL
		print FILE "\n**NEEDED_LIBS_IMPL\n";
		foreach my $Did ( sort keys %$all_distrs ) {
			my @libs = ();
			foreach my $soname ( @{$com->{$Did}{NEEDED_LIBS_IMPL}} ) {
				my $lib = $com->{$Did}{LIBS}{$soname};
				next if ( !$lib->{FOUND} || !$lib->{USED} );
				push @libs, $soname;
			}
			next if !@libs;
			print FILE $Did.":";
			print FILE join ",", @libs;
			print FILE "\n";
		}

		# found libraries
		print FILE "\n**FOUND_LIBS\n";
		foreach my $Did ( sort keys %$all_distrs ) {
			foreach my $soname ( sort keys %{$com->{$Did}{LIBS}} ) {
				my $lib = $com->{$Did}{LIBS}{$soname};
				next if ( !$lib->{FOUND} );
				print FILE $soname.":".$Did;
				print FILE ":F=".$lib->{FILE};
				print FILE ":DC=".$lib->{DISTR_COMPONENT} if $lib->{DISTR_COMPONENT};
				print FILE ":VIA=".$lib->{VIA} if $lib->{VIA};
				print FILE "\n";
			}
		}

		# needed interfaces
		print FILE "\n**NEEDED_INTS\n";
		foreach my $fullname ( sort keys %{$com->{NEEDED_INTS}} ) {
			print FILE $fullname."\n";
		}
		
		# found interfaces
		print FILE "\n**FOUND_INTS\n";
		my $values = {}; # group distr IDs by same values: $values->{$fullname}{$value} = [Distr IDs];
		foreach my $Did ( sort keys %$all_distrs ) {
			foreach my $fullname ( sort keys %{$com->{$Did}{FOUND_INTS}} ) {
				my $value = $com->{$Did}{FOUND_INTS}{$fullname}; # "$soname|$ver_found"
				
				push @{$values->{$fullname}{$value}}, $Did;
			}
		}
		foreach my $fullname ( sort keys %$values ) {
			foreach my $line ( sort keys %{$values->{$fullname}} ) {
				print FILE $fullname.":".(join ",", @{$values->{$fullname}{$line}});
				print FILE ":".$line;
				print FILE "\n";
			}
		}

		# MISSING_INTS
		print FILE "\n**MISSING_INTS\n";
		foreach my $Did ( sort keys %$all_distrs ) { # TODO: pack as for FOUND_INTS
			foreach my $fullname ( sort keys %{$com->{$Did}{MISSING_INTS}} ) {
				my $soname = $com->{$Did}{MISSING_INTS}{$fullname};
				print FILE $fullname.":".$Did.":".$soname;
				print FILE "\n";
			}
		}

		# LIBS_OTHER_VERSION
		print FILE "\n**LIBS_OTHER_VERSION\n";
		foreach my $Did ( sort keys %$all_distrs ) {
			foreach my $soname ( sort keys %{$com->{$Did}{LIBS_OTHER_VERSION}} ) {
				my $data = $com->{$Did}{LIBS_OTHER_VERSION}{$soname};
				print FILE $soname.":".$Did;
				print FILE ":S=".$data->{SONAME};
				print FILE ":F=".$data->{FILE};
				print FILE ":DC=".$data->{DISTR_COMPONENT};
				print FILE "\n";
			}
		}
	}

	print FILE "\n*COMMON_DATA\n";
	
	if ( $improper_arch ) {
		print FILE "IMPROPER_ARCH: $improper_arch\n";
	}
	
	# Common information about libraries
	print FILE "\n**LIBRARIES\n";
	my %tmp = ();
	foreach my $soname ( keys %$approved_libs ) {
		$tmp{$soname}{'AP'} = $approved_libs->{$soname};
	}
	foreach my $soname ( sort keys %tmp ) {
		print FILE join( ';', $soname, %{$tmp{$soname}} ), "\n";
	}

	# Common information about interfaces
	print FILE "\n**INTERFACES\n";
	%tmp = ();
	foreach my $Iname ( sort keys %$unmangled ) {
		$tmp{$Iname}{'UM'} = $unmangled->{$Iname};
	}
	foreach my $Iname ( sort keys %$unknown_ints ) {
		$tmp{$Iname}{'U'} = $unknown_ints->{$Iname};
	}
	foreach my $Iname ( sort keys %tmp ) {
		print FILE join( ';', $Iname, %{$tmp{$Iname}} ), "\n";
	}

	# LSB libraries
	print FILE "\n**LSB_LIBRARIES\n";
	foreach my $soname ( sort keys %$lsb_lib_status ) {
		my $lsb_lib = $lsb_lib_status->{$soname};
		print FILE $soname.":";
		print FILE "S=".$lsb_lib->{STATUS}.":";
		print FILE "V=".$lsb_lib->{STATUS_VER}.":" if defined $lsb_lib->{STATUS_VER};
		print FILE "R=".$lsb_lib->{REASON}.":" if defined $lsb_lib->{REASON};
		print FILE "\n";
	}

	# LibAttrs
	print FILE "\n**LIB_ATTRS\n";
	foreach my $soname ( sort keys %$lib_attrs ) {
		foreach my $LAtype ( sort keys %{$lib_attrs->{$soname}} ) {
			my $LAvalue = $lib_attrs->{$soname}{$LAtype};

			$LAvalue =~ s/[\n\r]/\x81/g; # Use \x81 instead of \n

			print FILE join("\t", $soname, $LAtype, $LAvalue);
			print FILE "\n";
		}
	}

	# LSB interfaces
	print FILE "\n**LSB_INTERFACES\n";
	foreach my $fullname ( sort keys %$lsb_int_status ) {
	foreach my $soname ( sort keys %{$lsb_int_status->{$fullname}} ) {
		my $lsb_int = $lsb_int_status->{$fullname}{$soname};
		print FILE $fullname.":".$soname.":";
		print FILE "S=".$lsb_int->{STATUS}.":";
		print FILE "V=".$lsb_int->{STATUS_VER}.":" if defined $lsb_int->{STATUS_VER};
		print FILE "R=".$lsb_int->{REASON}.":" if defined $lsb_int->{REASON};
		print FILE "\n";
	}}

	# IntAttrs
	print FILE "\n**INT_ATTRS\n";
	foreach my $Iname ( sort keys %$int_attrs ) {
		foreach my $lib ( sort keys %{$int_attrs->{$Iname}} ) {
			foreach my $LAtype ( sort keys %{$int_attrs->{$Iname}{$lib}} ) {
				my $LAvalue = $int_attrs->{$Iname}{$lib}{$LAtype};

				$LAvalue =~ s/[\n\r]/\x81/g; # Use \x81 instead of \n

				print FILE join( "\t", $Iname, $lib, $LAtype, $LAvalue );
				print FILE "\n";
			}
		}
	}

	close FILE;
}
#-----------------------------------------------------------------------

sub load_main {
	my ($filename) = @_;

	open FILE, $filename
		or return "Failed to open file for reading '$filename': $!";

	my $mode = "";
	my $com;
	LINES_LOOP:
	while ( my $line = <FILE> ) {
		next if $line =~ /^\s*$/; # Skip empty lines
		next if $line =~ /^#/; # Skip commented lines
		chomp $line;

		if ( $line =~ /^\*/ ) { # New mode
			foreach my $case ( qw/APP_COMPONENT NEEDED_LIBS NEEDED_LIBS_IMPL NEEDED_INTS
				FOUND_LIBS LIBS_OTHER_VERSION FOUND_INTS MISSING_INTS COMMON_DATA
				LIBRARIES INTERFACES LSB_LIBRARIES LIB_ATTRS LSB_INTERFACES INT_ATTRS/ )
			{
				if ( $line =~ /^\*+$case\b/ ) {
					$mode = $case;
					if ( $mode eq 'APP_COMPONENT' ) { # New application component
						if ( $com && !$com->{NAME} ) {
							return "Unsaved component: ".join(" ",%$com)
						};
						$com = { NEEDED_LIBS => [], NEEDED_INTS => {} };
					}
					next LINES_LOOP;
				}
			}
		}

		if ( $mode eq 'APP_COMPONENT' ) {
			if ( $line =~ /^([^:]+):\s*(.*)/ ) {
				my ($key, $value) = ($1, $2);
				$com->{$key} = $value;
				if ( $key eq 'NAME' ) { # save the component
					$app_components->{$value} = $com;
				}
			}
			else {
				return "Can't parse '$line' in file '$filename' (mode '$mode').";
			}
		}
		elsif ( $mode eq 'NEEDED_LIBS' ) {
			my $soname = $line;
			push @{$com->{NEEDED_LIBS}}, $soname;
		}
		elsif ( $mode eq 'NEEDED_LIBS_IMPL' ) {
			my ($Did, $sonames) = split /:/, $line, 2;
			$sonames ||= "";
			$com->{$Did}{NEEDED_LIBS_IMPL} = [ split /,/, $sonames ];
		}
		elsif ( $mode eq 'NEEDED_INTS' ) {
			my $fullname = $line;
			$com->{NEEDED_INTS}{$fullname} = 1;
		}
		elsif ( $mode eq 'FOUND_LIBS' ) {
			my ($soname, $Did, @pairs) = split /:/, $line;
			my $lib = $com->{$Did}{LIBS}{$soname} ||= { FOUND => 1 };
			foreach my $pair ( @pairs ) {
				my ($key, $val) = split /=/, $pair, 2;
				
				   if ( $key eq 'F' ) { $lib->{FILE} = $val; }
				elsif ( $key eq 'DC' ) { $lib->{DISTR_COMPONENT} = $val; }
				elsif ( $key eq 'VIA' ) { $lib->{VIA} = $val; }
			}
		}
		elsif ( $mode eq 'FOUND_INTS' ) {
			my ($fullname, $distrs, $data) = split /:/, $line;
			foreach my $Did ( split /,/, $distrs ) {
				$com->{$Did}{FOUND_INTS}{$fullname} = $data;
			}
		}
		elsif ( $mode eq 'MISSING_INTS' ) {
			my ($fullname, $distrs, $soname) = split /:/, $line;
			foreach my $Did ( split /,/, $distrs ) {
				$com->{$Did}{MISSING_INTS}{$fullname} = $soname;
			}
		}
		elsif ( $mode eq 'LIBS_OTHER_VERSION' ) {
			my ($soname_req, $Did, @pairs) = split /:/, $line;
			my $obj = $com->{$Did}{LIBS_OTHER_VERSION}{$soname_req} ||= {};
			foreach my $pair ( @pairs ) {
				my ($key, $val) = split /=/, $pair, 2;
				   if ( $key eq 'S' ) { $obj->{SONAME} = $val; }
				elsif ( $key eq 'F' ) { $obj->{FILE} = $val; }
				elsif ( $key eq 'DC' ) { $obj->{DISTR_COMPONENT} = $val; }
			}
		}
		elsif ( $mode eq 'COMMON_DATA' ) {
			if ( $line =~ /^\s*IMPROPER_ARCH:\s*(.*)/ ) {
				$improper_arch = $1;
			}
		}
		elsif ( $mode eq 'LIBRARIES' ) {
			my ($soname, @tmp) = split /\;/, $line;
			while ( @tmp ) {
				my $key = shift @tmp;
				my $val = shift @tmp;
				if ( $key eq 'AP' ) { $approved_libs->{$soname} = $val; }
			}
		}
		elsif ( $mode eq 'INTERFACES' ) {
			my ($Iname, @tmp) = split /\;/, $line;
			while ( @tmp ) {
				my $key = shift @tmp;
				my $val = shift @tmp;
				if ( $key eq 'UM' ) {
					$unmangled->{$Iname} = $val;
				}
				elsif ( $key eq 'U' ) {
					$unknown_ints->{$Iname} = $val;
				}
			}
		}
		elsif ( $mode eq 'LSB_LIBRARIES' ) {
			my ($soname, @pairs) = split /:/, $line;
			my $lsb_lib = $lsb_lib_status->{$soname} ||= {};
			foreach my $pair ( @pairs ) {
				my ($key, $val) = split /=/, $pair, 2;
				
				   if ( $key eq 'S' ) { $lsb_lib->{STATUS} = $val; }
				elsif ( $key eq 'V' ) { $lsb_lib->{STATUS_VER} = $val; }
				elsif ( $key eq 'R' ) { $lsb_lib->{REASON} = $val; }
			}
		}
		elsif ( $mode eq 'LIB_ATTRS' ) {
			my ($soname, $LAtype, $LAvalue) = split /\t/, $line;

			$LAvalue =~ s/\x81/\n/g; # Use \x81 instead of \n

			$lib_attrs->{$soname}{$LAtype} = $LAvalue;
		}
		elsif ( $mode eq 'LSB_INTERFACES' ) {
			my ($fullname, $soname, @pairs) = split /:/, $line;
			my $lsb_int = $lsb_int_status->{$fullname}{$soname} ||= {};
			foreach my $pair ( @pairs ) {
				my ($key, $val) = split /=/, $pair, 2;
				
				   if ( $key eq 'S' ) { $lsb_int->{STATUS} = $val; }
				elsif ( $key eq 'V' ) { $lsb_int->{STATUS_VER} = $val; }
				elsif ( $key eq 'R' ) { $lsb_int->{REASON} = $val; }
			}
		}
		elsif ( $mode eq 'INT_ATTRS' ) {
			my ($Iname, $lib, $LAtype, $LAvalue) = split /\t/, $line;

			$LAvalue =~ s/\x81/\n/g; # Use \x81 instead of \n

			$int_attrs->{$Iname}{$lib}{$LAtype} = $LAvalue;
		}
		elsif ( $mode eq '' ) {
			if ( $line =~ /^\s*FORMAT:\s*(\S+)/ ) {
				my $format_ver = $1;
				if ( $format_ver ne $save_format_ver ) {
					return "Unsupported format of the results file. Please test the application again.";
				}
			}
		}
		else {
			# ignore unknown modes
		}
	}

	close FILE;

	return; # OK
}
#-----------------------------------------------------------------------

sub complete_structures {
	my $ints_used = 0;
	
	# Fill in missed libs
	foreach my $com ( values %$app_components ) {
		foreach my $Did ( keys %$all_distrs ) {
			$com->{$Did}{NEEDED_LIBS_IMPL} ||= [];
			foreach my $soname ( @{$com->{NEEDED_LIBS}}, @{$com->{$Did}{NEEDED_LIBS_IMPL}} ) {
				$com->{$Did}{LIBS}{$soname}{INTERFACES} ||= 0;
			}
		}
	}
	
	# Add an interface list for each found library
	foreach my $com ( values %$app_components ) {
		foreach my $Did ( keys %$all_distrs ) {
			my $com_distr = $com->{$Did};
			my $com_distr_libs = $com_distr->{LIBS};
			foreach my $fullname ( keys %{$com_distr->{FOUND_INTS}} ) {
				my $soname = $com_distr->{FOUND_INTS}{$fullname};
				( $soname ) or die; # DBG
				$soname =~ s/\|.*//; # get soname
				
				my $lib = $com_distr_libs->{$soname}; # should be defined
				if ( $lib->{VIA} ) { $soname = $lib->{VIA}; }
				
				$com_distr_libs->{$soname}{INTERFACES} += 1;
			}
			foreach my $fullname ( keys %{$com_distr->{MISSING_INTS}} ) {
				my $soname = $com_distr->{MISSING_INTS}{$fullname};
				
				my $lib = $com_distr_libs->{$soname}; # may be undef
				if ( $lib && $lib->{VIA} ) { $soname = $lib->{VIA}; }
				
				$com_distr_libs->{$soname}{INTERFACES} += 1;
			}
		}
	}

	return 1; # OK
}
#-----------------------------------------------------------------------

sub unneeded_libs {
	
	# About the UNNEEDED status!
	#
	# Library is treated as unneeded if no interfaces is taken from this library
	# by ANY COMPONENT on ANY DISTRIBUTION, because we can't say that library is unneeded
	# if there is a component that requires this library on some distro.
	#
	# This is true for D_table, => Summary, distr_details, => com_distr_details, L_table.
	#
	# However for a certain application we can say, that the library
	# is unneeded if the component doesn't use any interface from this library
	# on ANY DISTRIBUTION without regard to other application components.
	#
	# This is true for C_table, => com_details, L_table filtered by component.
	#
	
	my $all_libs = {};
	
	# Library status
	foreach my $com ( keys %$app_components ) {
		my $com_obj = $app_components->{$com};
		foreach my $soname ( @{$com_obj->{NEEDED_LIBS}} ) {
			my $unneeded = 1;
			foreach my $Did ( keys %$all_distrs ) {
				my $lib_obj = $com_obj->{$Did}{LIBS}{$soname} or next;
				if ( $lib_obj->{INTERFACES} ) { # if there any interfaces used
					$unneeded = 0;
					last;
				}
			}
			$com_obj->{LIBS_UNNEEDED}{$soname} = $unneeded;
			$all_libs->{$soname} = 1;
		}
	}
	
	foreach my $soname ( keys %$all_libs ) {
		my $unneeded = 1;
		foreach my $com ( keys %$app_components ) {
			if ( defined $app_components->{$com}{LIBS_UNNEEDED}{$soname}
				&& !$app_components->{$com}{LIBS_UNNEEDED}{$soname} ) 
			{
				$unneeded = 0;
				last;
			}
		}
		$libs_unneeded->{$soname} = 1 if $unneeded;
	}
}

# Gather the data on a given component and distribution
sub com_distr_data {
	my ($com_obj, $Did, $unneeded_libs, $libs) = @_;
	
	my $com_distr = $com_obj->{$Did};
	
	# Library status
	foreach my $soname ( @{$com_obj->{NEEDED_LIBS}}, ($libs ? @$libs : ()) ) {
		if ( $com_distr->{LIBS}{$soname}{FOUND} ) {
			# Present
			$com_distr->{LIBS}{$soname}{STATUS} = 'OK';
		}
		else { # Missing
			if ( ! $approved_libs->{$soname} ) {
				# Unknown lib
				$com_distr->{LIBS}{$soname}{STATUS} = 'UNKNOWN';
			}
			elsif ( $unneeded_libs->{$soname} ) {
				# Unneeded lib
				$com_distr->{LIBS}{$soname}{STATUS} = 'UNNEEDED';
			}
			else {
				# Known missing lib
				$com_distr->{LIBS}{$soname}{STATUS} = 'MISSING';
			}
		}
	}

	# Interface statistics
	foreach my $fullname ( keys %{$com_obj->{NEEDED_INTS}} ) {
		if ( $com_distr->{FOUND_INTS}{$fullname} ) {
			# Present
			$com_distr->{INT_STATUS}{$fullname} = 'OK';
		}
		else { # Missing
			my $soname = $com_distr->{MISSING_INTS}{$fullname};
			if ( defined $soname ) { # linked
				if ( $approved_libs->{$soname} ) {
					# Known missing int
					$com_distr->{INT_STATUS}{$fullname} = 'MISSING';
				} else {
					# Unapproved library
					$com_distr->{INT_STATUS}{$fullname} = 'UNKNOWN';
				}
			}
			else { # unlinked
				my $Iname = $fullname; $Iname =~ s/\@.*//;
				if ( $Analysis::unknown_ints->{$Iname} ) {
					# Unknown int
					$com_distr->{INT_STATUS}{$fullname} = 'UNKNOWN';
				} else {
					# Missing int (non-linked)
					$com_distr->{INT_STATUS}{$fullname} = 'MISSING';
				}
			}
		}
	}
}

# Requires com_distr_data to be called before
sub com_distr_status {
	my ($com_obj, $Did) = @_;
	
	my $com_distr = $com_obj->{$Did};
	
	return $com_distr->{STATUS} if defined $com_distr->{STATUS}; # already calculated

	# Sum up the status:
	my $stat = {};
	@$stat{qw/MISSING_LIBS UNNEEDED_LIBS UNKNOWN_LIBS MISSING_INTS UNKNOWN_INTS/} = (0) x 5;
	foreach my $req_soname ( @{$com_obj->{NEEDED_LIBS}} ) {
		my $val = $com_distr->{LIBS}{$req_soname}{STATUS};
		if ( $val eq 'MISSING' ) { $stat->{MISSING_LIBS}++; }
		if ( $val eq 'UNNEEDED' ) { $stat->{UNNEEDED_LIBS}++; }
		if ( $val eq 'UNKNOWN' ) { $stat->{UNKNOWN_LIBS}++; }
	}
	foreach my $val ( values %{$com_distr->{INT_STATUS}} ) {
		if ( $val eq 'MISSING' ) { $stat->{MISSING_INTS}++ }
		if ( $val eq 'UNKNOWN' ) { $stat->{UNKNOWN_INTS}++; }
	}

	# General status
	if ( $stat->{MISSING_LIBS} + $stat->{MISSING_INTS} ) {
		$com_distr->{STATUS} = 'MISSING';
	}
	elsif ( $stat->{UNKNOWN_LIBS} + $stat->{UNKNOWN_INTS} ) {
		$com_distr->{STATUS} = 'UNKNOWN';
	}
	elsif ( $stat->{UNNEEDED_LIBS} ) {
		$com_distr->{STATUS} = 'UNNEEDED';
	}
    else {
        $com_distr->{STATUS} = 'OK';
    }

	# Additionaly, check the .hash section type
	if ( $com_obj->{NO_SHT_HASH} && $all_distrs->{$Did}{NO_GNU_HASH} ) {
		$com_distr->{STATUS} = 'MISSING';
	}
	
	return $com_distr->{STATUS};
}
#-----------------------------------------------------------------------

# Calculate the status of a given distribution
# This function is used for calculation of a final verdict
our %status_range = ('' => 0, 'OK' => 1, 'UNNEEDED' => 2, 'UNKNOWN' => 3, 'MISSING' => 4); # also used in D_table for sorting
sub distr_status {
	my ($Did) = @_;

	my $distr = $all_distrs->{$Did};

	my $status = '';
	
	foreach my $com ( keys %$app_components ) {
		my $com_obj = $app_components->{$com};
		
		com_distr_data( $com_obj, $Did, $libs_unneeded );

		# Sum up the overall status
		my $com_d_status = com_distr_status( $com_obj, $Did );
		
		if ( $status_range{$com_d_status} > $status_range{$status} ) {
			$status = $com_d_status;
		}
	}
	
	return $status;
}
#-----------------------------------------------------------------------

# Gather the data about a given distribution
sub distr_data {
	my ($Did) = @_;

	my $distr = $all_distrs->{$Did};
	return undef if !$distr;

	# return if already calculated
	if ( $distr->{STATUS} && $distr->{LIBS} && $distr->{INTS} ) { return $distr; }
	
	# Add
	my $distr_status = distr_status($Did);
	$distr->{STATUS} = $distr_status;
	$distr_status = undef; # delete

	# Determine the status of each library and interface
	$distr->{LIBS} = {};
	$distr->{INTS} = {};
	foreach my $com_obj ( values %$app_components ) {
		foreach my $soname ( @{$com_obj->{NEEDED_LIBS}} ) {
			my $status_val = $com_obj->{$Did}{LIBS}{$soname}{STATUS};
			my $prev_val = $distr->{LIBS}{$soname};
			if ( !$prev_val || $status_range{$status_val} > $status_range{$prev_val} ) {
				$distr->{LIBS}{$soname} = $status_val;
			}
		}
		my $ints = $com_obj->{$Did}{INT_STATUS};
		foreach my $fullname ( keys %$ints ) {
			my $status_val = $ints->{$fullname};
			my $prev_val = $distr->{INTS}{$fullname};
			if ( !$prev_val || $status_range{$status_val} > $status_range{$prev_val} ) {
				$distr->{INTS}{$fullname} = $ints->{$fullname};
			}
		}

		if ( $com_obj->{NO_SHT_HASH} && $distr->{NO_GNU_HASH} ) {
			$distr->{COMMENTS}{NO_GNU_HASH} = 1;
		}
	}
	
	@$distr{qw/MISSING_LIBS UNNEEDED_LIBS UNKNOWN_LIBS MISSING_INTS UNKNOWN_INTS/} = (0) x 5;
	foreach my $val ( values %{$distr->{LIBS}} ) {
		if ( $val eq 'MISSING' ) { $distr->{MISSING_LIBS}++ }
		if ( $val eq 'UNNEEDED' ) { $distr->{UNNEEDED_LIBS}++; }
		if ( $val eq 'UNKNOWN' ) { $distr->{UNKNOWN_LIBS}++; }
		if ( $val eq 'OK' ) { $distr->{OK_LIBS}++; }
	}
	foreach my $val ( values %{$distr->{INTS}} ) {
		if ( $val eq 'MISSING' ) { $distr->{MISSING_INTS}++ }
		if ( $val eq 'UNKNOWN' ) { $distr->{UNKNOWN_INTS}++; }
		if ( $val eq 'OK' ) { $distr->{OK_INTS}++; }
	}

	return $distr;
}
#-----------------------------------------------------------------------

# Gather information about the given component
sub com_data {
	my ($com_obj) = @_;  # TODO: use distr_sel
	
	# return if already calculated
	if ( $com_obj->{LIB_STATUS} ) { return; }
	
	$com_obj->{LIB_STATUS} = {};
	
	# Determine the status of each library and interface
	foreach my $Did ( keys %$all_distrs ) {
		com_distr_data( $com_obj, $Did, $com_obj->{LIBS_UNNEEDED} );
		
		foreach my $soname ( @{$com_obj->{NEEDED_LIBS}} ) {
			my $com_distr_lib = $com_obj->{$Did}{LIBS}{$soname};
			my $status_val = $com_distr_lib->{STATUS};
			
			my $prev_val = $com_obj->{LIB_STATUS}{$soname};
			if ( !$prev_val || $status_range{$status_val} > $status_range{$prev_val} ) { # choose the worst case
				$com_obj->{LIB_STATUS}{$soname} = $status_val;
			}
		}
		
		# Use the NEEDED_INTS hash to store interface statuses.
		foreach my $fullname ( keys %{$com_obj->{$Did}{INT_STATUS}} ) {
			my $status_val = $com_obj->{$Did}{INT_STATUS}{$fullname};
			
			# store the interface status directly in the NEEDED_INTS field
			my $prev_val = $com_obj->{NEEDED_INTS}{$fullname};
			$prev_val = '' if $prev_val eq 1; # by default there was 1
			if ( !$prev_val || $status_range{$status_val} > $status_range{$prev_val} ) { # choose the worst case
				$com_obj->{NEEDED_INTS}{$fullname} = $status_val;
			}
		}
	}
}
#-----------------------------------------------------------------------

my %cmp_lsb_status_order = (NOT=>0, WITHDRAWN=>1, CANDIDATE=>2, FUTURE=>3,
		 DEPRECATED=>4, OPTIONAL=>5, IN=>6 );

sub cmp_lsb_status {
	my ($a, $b) = @_;
	
	if ( !$a && !$b ) { return 0; }
	if ( !$a ) { return -1; }
	if ( !$b ) { return 1; }
	
	my $tmp = ( $cmp_lsb_status_order{$a->{STATUS}} <=> $cmp_lsb_status_order{$b->{STATUS}} );
	return $tmp if $tmp;
	
	if ( $a->{STATUS_VER} && $b->{STATUS_VER} ) {
		my $status = $a->{STATUS};
		if ( $status eq 'DEPRECATED' || $status eq 'WITHDRAWN' ) {
			$tmp = +($a->{STATUS_VER} cmp $b->{STATUS_VER});
		} else {
			$tmp = -($a->{STATUS_VER} cmp $b->{STATUS_VER});
		}
	}
	return $tmp if $tmp;
	
	return 0;
}

## Calculate the LSB status for a given component
sub com_data_lsb {
	my ($com_obj) = @_;
	
	# Libs
	foreach my $soname ( @{$com_obj->{NEEDED_LIBS}} ) {
		$com_obj->{LSB_STATUS}{LIBS}{$soname} = $lsb_lib_status->{$soname};
	}
	
	foreach my $fullname ( keys %{$com_obj->{NEEDED_INTS}} ) {
		# Gather different cases of libraries the interface is linked to.
		my %linked_libs = ();
		foreach my $Did ( keys %$all_distrs ) {
			my $soname = $com_obj->{$Did}{FOUND_INTS}{$fullname};
			if ( $soname ) { $soname =~ s/\|.*// }
			$soname ||= $com_obj->{$Did}{MISSING_INTS}{$fullname};
			$soname ||= '';
			
			my $lib_obj = $com_obj->{$Did}{LIBS}{$soname};
			if ( $lib_obj && $lib_obj->{VIA} ) { $soname = $lib_obj->{VIA}; }
			
			$linked_libs{$soname} = 1;
		}
		
		# Summarizing LSB status of interfaces over distributions
		my $status_obj;
		for my $soname ( keys %linked_libs ) {
			my $status_obj_tmp = $lsb_int_status->{$fullname}{$soname};
			
			if ( !$status_obj || cmp_lsb_status($status_obj_tmp, $status_obj) < 0 ) {
				$status_obj = $status_obj_tmp;
			}
		}
		$com_obj->{LSB_STATUS}{INTS}{$fullname} = $status_obj;
	}
}
#-----------------------------------------------------------------------

sub com_common_data {
	my ($com_obj) = @_;
	
	my $soname = $com_obj->{SONAME};
	
	if ( !$soname ) { return; }
	if ( !$approved_libs->{$soname} ) { return; }
	
	$com_obj->{COMMON} = 1;
	$com_obj->{COMMON_DISTRS} = {};
	
	foreach my $Did ( keys %$all_distrs ) {
		if ( $com_obj->{$Did}{LIBS}{$soname}
			&& $com_obj->{$Did}{LIBS}{$soname}{FOUND} )
		{
			$com_obj->{COMMON_DISTRS}{$Did} = 1;
		}
	}
	
	return $com_obj;
}
#-----------------------------------------------------------------------

# Returns LSB status of the library and LSB version related to this status
# Values: NOT CANDIDATE FUTURE IN OPTIONAL DEPRECATED WITHDRAWN
sub lsb_status {
	my ($obj, $selected_lsb_ver) = @_;

	my %events = ();
	
	if ( $obj->{APPEAREDIN} ) {
		$events{$_} = 'APPEARED' foreach ( split /,\s*/, $obj->{APPEAREDIN} );
	}
	if ( $obj->{DEPRECATEDSINCE} ) {
		$events{$_} = 'DEPRECATED' foreach ( split /,\s*/, $obj->{DEPRECATEDSINCE} );
	}
	if ( $obj->{WITHDRAWNIN} ) {
		$events{$_} = 'WITHDRAWN' foreach ( split /,\s*/, $obj->{WITHDRAWNIN} );
	}
	
	my $cur_state = 'NOT';
	my $last_v = '';
	foreach my $v ( sort keys %events ) {
		last if $v gt $selected_lsb_ver;

		$cur_state = $events{$v};
		$last_v = $v;
	}
	
	if ( $cur_state eq 'NOT' ) {
		if ( $obj->{APPEAREDIN} ) {
			my $first_appear = (sort split /,\s*/, $obj->{APPEAREDIN})[0];
			return {STATUS => "FUTURE", STATUS_VER => $first_appear};
		}
		elsif ( $obj->{CANDIDATEFOR} ) {
			return {STATUS => "CANDIDATE", STATUS_VER => $obj->{CANDIDATEFOR} };
		}
		else {
			return {STATUS => "NOT"};
		}
	}
	elsif ( $cur_state eq 'APPEARED' ) {
		if ( $obj->{MANDATORYSINCE} && $obj->{MANDATORYSINCE} le $selected_lsb_ver ) {
			return {STATUS => "IN", STATUS_VER => $last_v};
		} else {
			return {STATUS => "OPTIONAL", STATUS_VER => $last_v};
		}
	}
	elsif ( $cur_state eq 'DEPRECATED' ) {
		return {STATUS => "DEPRECATED", STATUS_VER => $last_v};
	}
	elsif ( $cur_state eq 'WITHDRAWN' ) {
		return {STATUS => "WITHDRAWN", STATUS_VER => $last_v};
	}
}

sub lib_lsb_status {
	my ($soname, $selected_lsb_ver) = @_;
	
	die if !$soname;
	
	my $lsb_status = undef;
	
	# Look up the LSB library name for $soname
	my $lsb_lib_obj = undef;
	if ( $lsb_libs->{$soname} ) {
		$lsb_lib_obj = $lsb_libs->{$soname};
		$lsb_status = lsb_status($lsb_lib_obj, $selected_lsb_ver);
	}
	if ( !$lsb_lib_obj ) { # not found
		# TODO: wrong ver?
	}
	
	if ( !$lsb_lib_obj ) { # not found
		$lsb_status = { STATUS=>'NOT' };
	}
	
	return $lsb_status;
}

sub calc_lsb_lib_status {
	my ($selected_lsb_ver) = @_;
	
	foreach my $com_obj ( values %$app_components ) {
		foreach my $soname ( @{$com_obj->{NEEDED_LIBS}} ) {
			next if $lsb_lib_status->{$soname};
			$lsb_lib_status->{$soname} = lib_lsb_status($soname, $selected_lsb_ver);
		}
	}
}

sub int_lsb_status {
	my ($Iname, $ver, $soname, $selected_lsb_ver) = @_;
	
	my $lsb_status = undef;
	
	# Look up the LSB library name for $soname
	my $libname = undef;
	if ( $soname && $lsb_libs->{$soname} ) {
		$libname = $lsb_libs->{$soname}{LSBNAME};
	}
	
	$lsb_status = { STATUS => 'NOT' };
	
	# Search for lsb_int_obj
	my $lsb_int_obj = undef;
	if ( $lsb_ints->{$Iname} ) { # Check the name
		if ( $lsb_ints->{$Iname}{$ver} ) { # Check the version
			if ( $libname ) { # If the library is known
				if ( $lsb_ints->{$Iname}{$ver}{$libname} ) { # Check the library
					$lsb_int_obj = $lsb_ints->{$Iname}{$ver}{$libname};
					$lsb_status = lsb_status($lsb_int_obj, $selected_lsb_ver);
				} else {
					$lsb_status->{REASON} = 'WRONG_LIB';
					# TODO: select the proper lib
				}
			} else {
				if ( $soname ) {
					# TODO: non-lsb library; suggest another library?
				}
				# else: non-linked -- do nothing
			}
		} else {
			# TODO: select the best version
			# if there is an LSB interface with another version
			#	 then: $lsb_status->{REASON} = 'WRONG_VER';
		}
	}
	
	if ( $lsb_status->{STATUS} eq 'IN' ) {
		if ( $soname && $lsb_lib_status->{$soname} ) {
			# Check if the library is deprecated or withdrawn
			if ( $lsb_lib_status->{$soname}{STATUS} eq 'DEPRECATED' ) {
				$lsb_status = { STATUS=>'DEPRECATED',
						STATUS_VER => $lsb_lib_status->{$soname}{STATUS_VER},
					};
			}
			if ( $lsb_lib_status->{$soname}{STATUS} eq 'WITHDRAWN' ) {
				$lsb_status = { STATUS=>'WITHDRAWN',
						STATUS_VER => $lsb_lib_status->{$soname}{STATUS_VER},
					};
			}
		}
	}
	
	return $lsb_status;
}

sub calc_lsb_int_status {
	my ($selected_lsb_ver) = @_;
	
	foreach my $com_obj ( values %$app_components ) {
		foreach my $Did ( keys %$all_distrs ) {
			foreach my $fullname ( keys %{$com_obj->{NEEDED_INTS}} ) {
				my $soname = $com_obj->{$Did}{FOUND_INTS}{$fullname};
				if ( $soname ) { $soname =~ s/\|.*// }
				$soname ||= $com_obj->{$Did}{MISSING_INTS}{$fullname};
				$soname ||= '';
				
				next if $lsb_int_status->{$fullname}{$soname}; # already calculated
		
				my ($Iname, $ver) = split /\@/, $fullname, 2;
				$ver ||= "";
				
				$lsb_int_status->{$fullname}{$soname} =
					 int_lsb_status($Iname, $ver, $soname, $selected_lsb_ver);
			}
		}
	}
}

#-----------------------------------------------------------------------

sub overall_verdict {
	my $statistics;
	foreach my $Did (keys %$Analysis::all_distrs) {
		my $status = Analysis::distr_status($Did);
		$statistics->{$status}++;
	}

	my $result = '';
	if ( $statistics->{UNKNOWN} || $statistics->{UNNEEDED} # unknown or unneeded
		 || $statistics->{OK} && $statistics->{MISSING} ) # or mixed status
	{
		$result = 'Warnings';
	}
	elsif ( $statistics->{OK} ) { # only OK
		$result = 'Success';
	}
	elsif ( $statistics->{MISSING} ) { # only MISSING
		$result = 'Failed';
	}
	else {
		# No binaries
		$result = 'Warnings';
	}
	
	# Check improper arch:
	if ( $Analysis::improper_arch && $result eq 'Success' ) {
		$result = 'Warnings';
	}
	
	return $result;
}

1;
