#!/usr/bin/perl -w

# Linux Application Checker
# Configuration Save Module (tests_conf_save.pl)
#
# Copyright (C) 2007-2008 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use Cwd qw(cwd);
use Fcntl ':flock';
use Fcntl ':seek';
use File::Temp qw/tempfile tmpnam/;

sub BEGIN {
	my $utils_path = $0;
	$utils_path =~ s!/[^/]+$!!;
	if ($utils_path !~ m!^/!) {
		my $cwd = `pwd`;
		$cwd =~ s/[\r\n]//g;
		$utils_path = "$cwd/$utils_path";
	}
	push @INC, $utils_path;
}

use Common;

autoflush_on();

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# STARTUP SECTION

# Prints information about command-line parameters
sub usage() {
	print <<DATA;
    Usage: $0 [OPTION]...
    At least one of the --profile, --paths and --packages must be specified!

    --paths=<paths-list>     List of files and directories to be tested. For
                             directories all their contents is taken. The list
                             is delimited by a colon ":".
    --packages=<packages-list>
                             List of installed packages (rpm or deb) which
                             contents is to be tested. The list is delimited by
                             a colon ":".
    -n <name>, --name=<name> Name of the application. If not specified, the name
                             of the first file/directory/package will be taken.
    --lsb=<LSB-version>      LSB version to test against (4.0 by default).
    -T [core,c++|core,c++,desktop]
                             LSB profile to test against ("Core & C++ & Desktop"
                             by default).
    --comment=<text>         Comments to the test run. Several comments being
                             provided, they will be joined.
    -f <file>, --profile=<file>
                             Alternatively to specifying all the options above,
                             you may provide a text file containing all of them
                             inside. The profile can be created beforehand by
                             means of the Linux App Checker Web-UI.
    -r <dir>, --results-dir=<dir>
                             Directory for storing the results. If not specified,
                             "../results" will be used.
    -t <dir>, --tests-dir=<dir>
                             Directory containing lsbappchk and lsbpkgchk. If not
                             specified, "../bin" will be used.
    -m [rpm|dpkg]            Force using the specified package manager (RPM or
                             dpkg).
    --verbose                Output information about test stages.
    -h, --help               Display this help and exit.
    -v, --version            Output version information and exit.
DATA
}

my $error;

my $utils_path = abs_path($0);                  # 'utils' path
$utils_path =~ s!/[^/]+$!!;

# Parameters for testing (received via command line or with default values)
my @paths = ();                                 # Files/directories to be tested (--paths)
my @packages = ();                              # Packages to be tested (--packages)
my $app_name = '';                              # Application name (--name)
my $lsb_version = '4.0';                        # LSB version (--lsb)
my $lsb_version_specified = 0;
my $lsb_profile = 'core,c++,desktop';           # LSB profile (-T)
my $lsb_profile_specified = 0;
my $comments = '';                              # Comments to the test run (--comment)
my $profile = '';                               # Profile (--profile)
my $results_path = '';                          # 'results' directory (--results-dir)
my $results_path_specified = 0;
my $tests_path = '';                            # 'bin' directory (--tests-dir)
my $tests_path_specified = 0;
my $package_manager = '';                       # Package manager to be used (-m)
my $verbose = 0;                                # Output test stages (--verbose)
my $show_help = 0;                              # Display help (--help)
my $show_version = 0;                           # Display version (--version)

if ($utils_path eq '/opt/lsb/app-checker/utils') {
	$results_path = '/var/opt/lsb/app-checker/results';
	$tests_path = '/opt/lsb/bin';
}
else {
	$results_path = abs_path("$utils_path/../results");
	$tests_path = abs_path("$utils_path/../bin");
}

my $architecture = guess_test_arch("$tests_path/lsbappchk");    # LSB architecture name
my $f_architecture = lc($architecture);                         # File system architecture name

# Reads next command line parameter.
# Parameters:
#   1. (IN)  Ref to array of hashes containing info about options: {LONG_NAME => 'param', SHORT_NAME => 'p', HAS_VALUE => [1|0]}
#   2. (OUT) Ref to array with long name of the parameter (or short, if not present) and its value (undef if not needed).
#      If parameter is invalid, this parameter receives (undef, undef).
# Returns 1 if the parameter has been read successfully, 0 otherwise.
sub getoption($$) {
	my ($options, $result) = @_;
	my $key;
	my $val;
	my $param = shift @ARGV;
	if (!$param) {
		return 0;
	}
	foreach (@$options) {
		my $long_name = $_->{'LONG_NAME'};
		my $short_name = $_->{'SHORT_NAME'};
		my $return_name = ($long_name or $short_name);
		if ($_->{'HAS_VALUE'}) {
			if ($long_name and ($param =~ m/^--\Q$long_name\E=(.+)$/)) {
				@$result = ($return_name, $1);
				return 1;
			}
			elsif ($short_name) {
				if ($param =~ m/^-\Q$short_name\E(.+)$/) {
					@$result = ($return_name, $1);
					return 1;
				}
				elsif ($param =~ m/^-\Q$short_name\E$/) {
					my $value = shift @ARGV;
					if (!$value) {
						print "    Error: Option $param must be followed by a value!\n";
						$error = 1;
						@$result = (undef, undef);
					}
					else {
						@$result = ($return_name, $value);
					}
					return 1;
				}
			}
		}
		else {
			if ($long_name and ($param =~ m/^--\Q$long_name\E$/)) {
				@$result = ($return_name, undef);
				return 1;
			}
			elsif ($short_name and ($param =~ m/^-\Q$short_name\E$/)) {
				@$result = ($return_name, undef);
				return 1;
			}
		}
	}
	print "    Error: Unknown option: $param\n";
	$error = 1;
	@$result = (undef, undef);
	return 1;
}

my @cmd_line_options = (
	{'LONG_NAME' => 'paths', 'SHORT_NAME' => undef, 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'packages', 'SHORT_NAME' => undef, 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'name', 'SHORT_NAME' => 'n', 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'lsb', 'SHORT_NAME' => undef, 'HAS_VALUE' => 1},
	{'LONG_NAME' => undef, 'SHORT_NAME' => 'T', 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'comment', 'SHORT_NAME' => undef, 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'profile', 'SHORT_NAME' => 'f', 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'results-dir', 'SHORT_NAME' => 'r', 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'tests-dir', 'SHORT_NAME' => 't', 'HAS_VALUE' => 1},
	{'LONG_NAME' => undef, 'SHORT_NAME' => 'm', 'HAS_VALUE' => 1},
	{'LONG_NAME' => 'verbose', 'SHORT_NAME' => undef, 'HAS_VALUE' => 0},
	{'LONG_NAME' => 'help', 'SHORT_NAME' => 'h', 'HAS_VALUE' => 0},
	{'LONG_NAME' => 'version', 'SHORT_NAME' => 'v', 'HAS_VALUE' => 0}
);

# Process command line
my @param = ();
while (getoption(\@cmd_line_options, \@param)) {
	next if (!$param[0]);
	if ($param[0] eq 'paths') {
		@paths = map(abs_path($_), split(/:/, $param[1]));
	}
	elsif ($param[0] eq 'packages') {
		@packages = split(/:/, $param[1]);
	}
	elsif ($param[0] eq 'name') {
		$app_name = $param[1];
	}
	elsif ($param[0] eq 'lsb') {
		$lsb_version = $param[1];
		$lsb_version_specified = 1;
	}
	elsif ($param[0] eq 'T') {
		$lsb_profile = $param[1];
		$lsb_profile_specified = 1;
	}
	elsif ($param[0] eq 'comment') {
		$comments .= (($comments eq '') ? '' : "\n").$param[1];
	}
	elsif ($param[0] eq 'profile') {
		$profile = $param[1];
	}
	elsif ($param[0] eq 'results-dir') {
		$results_path = abs_path($param[1]);
		$results_path_specified = 1;
	}
	elsif ($param[0] eq 'tests-dir') {
		$tests_path = abs_path($param[1]);
		$tests_path_specified = 1;
	}
	elsif ($param[0] eq 'm') {
		$package_manager = $param[1];
	}
	elsif ($param[0] eq 'verbose') {
		$verbose = 1;
	}
	elsif ($param[0] eq 'help') {
		$show_help = 1;
	}
	elsif ($param[0] eq 'version') {
		$show_version = 1;
	}
}

if ($error) {                           # There were problems with command-line parameters
	print "    Run $0 --help for more information.\n";
	exit 1;
}

if ($show_help) {
	usage();
	exit 0;
}

if ($show_version) {
	my $ver_file = escape_path("$utils_path/VERSION");
	my $ver = (`cat $ver_file 2>/dev/null` or 'undefined');
	$ver =~ s/[\r\n]//g;
	print <<DATA;
Linux Application Checker version $ver
Copyright (C) 2007-2008 The Linux Foundation. All rights reserved.

DATA
	exit 0;
}

# If profile is specified, read it
if ($profile) {
	if (!open(FILE, $profile)) {
		print "    ERROR: Could not open file $profile\nERROR: $!\n";
		exit 2;
	}
	while (<FILE>) {
		s/[\r\n]//g;
		if (m/^APPLICATION:\s*(.*)/i) {
			$app_name = $1;
		}
		elsif (m/^COMPONENT:\s*(.*)/i) {
			my $comp = $1;
			if ($comp =~ /^pkg:(.*)/) {
				push @packages, $1;
			}
			else {
				push @paths, abs_path($comp);
			}
		}
		elsif (m/^ARCHITECTURE:\s*(.*)/i) {
			$architecture = $1;
			$f_architecture = lc($architecture);
		}
		elsif (m/^LSB_VERSION:\s*(.*)/i) {
			$lsb_version = $1;
		}
		elsif (m/^LSB_PROFILE:\s*(.*)/i) {
			$lsb_profile = $1;
		}
		elsif (m/^COMMENTS: (.*)/i) {
			$comments .= (($comments eq '') ? '' : "\n").$1;
		}
	}
	close(FILE);
}

# Check all parameters for consistency and fill in the default values when necessary
if (!$package_manager) {
	$package_manager = ($is_packager_rpm ? 'rpm' : 'dpkg');
}
else {
	if ((lc($package_manager) ne 'rpm') and (lc($package_manager) ne 'dpkg')) {
		print "    Error: Unknown package manager '$package_manager' specified!\n";
		print "    Run $0 --help for more information.\n";
		exit 1;
	}
}

if ((lc($lsb_profile) ne 'core,c++') and (lc($lsb_profile) ne 'core,c++,desktop')) {
	print "    Error: Unknown LSB profile '$lsb_profile' specified!\n";
	print "    Run $0 --help for more information.\n";
	exit 1;
}

if ((scalar(@paths) == 0) and (scalar(@packages) == 0)) {
	print "    Error: There is nothing to test!\n";
	print "    Run $0 --help for more information.\n";
	exit 1;
}

print "    Using LSB $lsb_version (".join(' & ', map(ucfirst, split(/,/, $lsb_profile))).")\n";
print "    Results directory is not specified, using $results_path\n" if (!$results_path_specified);
print "    Tests directory is not specified, using $tests_path\n" if (!$tests_path_specified);
print "\n";

if (!$app_name) {
	if (scalar(@paths) > 0) {
		$app_name = $paths[0];
		$app_name =~ m!/([^/]+)/?$!;
		$app_name = $1;
	}
	else {
		$app_name = $packages[0];
	}
}

# Escaped strings for using in system or `` calls
my $results_path_esc = escape_path($results_path);
my $tests_path_esc = escape_path($tests_path);
my $utils_path_esc = escape_path($utils_path);
my $app_name_esc = escape_path($app_name);


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# PREPARE DATA SECTION

my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime();
my $host_name = `uname -n`;
$host_name =~ s/[\r\n]//g;

# Directory name for this specific test run
my $results_dir = sprintf("$f_architecture-$host_name-$app_name-%04d-%02d-%02d-%02dh-%02dm-%02ds", ($year + 1900), ($mon + 1), $mday, $hour, $min, $sec);
my $results_dir_esc = escape_path($results_dir);

# Output file
my $tests_output = "$results_path/$results_dir_esc/test_logs/tests_output";

# Detects type of the file
sub detect_file_type($) {
	my ($path) = @_;
	if (!-e $path) {
		return 'error:File not found';
	}
	my $path_esc = escape_path($path);
	if (-d $path) {
		return 'dir';       # Directory
	}

	if (($path =~ m/\.tgz$/) or ($path =~ m/\.tar\.gz$/)) {
		return 'tgz';
	}
	elsif (($path =~ m/\.tbz$/) or ($path =~ m/\.tbz2$/) or ($path =~ m/\.tb2$/) or ($path =~ m/\.tar\.bz2$/)) {
		return 'tbz';
	}
	elsif ($path =~ m/\.tar$/) {
		return 'tar';
	}
	elsif ($path =~ m/\.rpm$/) {
		return 'rpm';
	}
	elsif ($path =~ m/\.deb$/) {
		return 'deb';
	}

	my $type = `POSIXLY_CORRECT=1 file $path_esc`;
	if ($type =~ m/\bELF\b/) {
		if ($type =~ m/\brelocatable\b/) {
			return 'elf-unsupported';                                           # Unsupported ELF type
		}
		else {
			return (($type =~ m/\bshared object\b/) ? 'elf-so' : 'elf-bin');    # Supported ELF type
		}
	}

	my $dft_err_file = tmpnam();
	my $first_line = `head -n1 $path_esc 2>$dft_err_file`;
	my $res = $?;
	my $err = `cat $dft_err_file`;
	unlink($dft_err_file);
	if (!$res) {
		if ($first_line =~ m/^#!/) {
			my $interpreter = '';
			if ($first_line =~ m|^#!\s*\S*/([^/\s]+)(\s.*)?$|) {
				$interpreter = lc($1);
				return 'perl' if ($interpreter eq 'perl');
				return 'python' if ($interpreter eq 'python');
				return 'shell' if ($interpreter =~ m/^(a|ba|da|z|k|c|tc|)sh$/);
			}
			return "script:$interpreter";   # Unknown script interpreter
		}
		else {
			return 'perl' if ($path =~ m/\.(pl|pm)$/);      # Perl module
			return 'python' if ($path =~ m/\.py$/);         # Python module
		}
		return 'unknown';
	}
	else {
		return "error:$err";
	}
}

# Constructs for the specific file a unique journal name compatible with lsbappchk's ones.
# Returns array: first element is the journal name generated, second - HTML anchor name for it.
sub get_journal_suffix_anchor($) {
	my ($fpath) = @_;
	if ($fpath =~ m/^PKG:\s*(.*)/) {
		my $pkg_name = $1;
		return ("$pkg_name.0-0.journal", "${pkg_name}_0_0");
	}
	else {
		my $fname = $fpath;
		$fname =~ s|^.*/([^/]*)$|$1|;
		my @tmp = stat($fpath);
		my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime, $mtime, $ctime, $blksize, $blocks) = ((scalar(@tmp) > 0) ? @tmp : (0)x13);
		my $fname_anchor = $fname;
		$fname_anchor =~ s/([^a-zA-Z0-9\-_])/sprintf('.%02X', ord($1))/eg;
		return ("$fname.$dev-$ino.journal", "${fname_anchor}_${dev}_${ino}");
	}
}

# Creates small journal just for indicating single PASS or FAIL.
sub create_journal($$$;$) {
	my ($path, $verdict, $type, $msg) = @_;
	my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime();
	$year += 1900;
	$mon++;
	my $tm = sprintf("%02d:%02d:%02d", $hour, $min, $sec);
	my $date = sprintf("%04d%02d%02d", $year, $mon, $mday);
	if ($msg) {
		$msg = join("\n", map {$_ = "520|0 1 $tm|$_"} split(/\n/, $msg));
	}
	my $result = ($verdict ? "$msg\n220|0 1 $verdict $tm|FAIL" : "220|0 1 0 $tm|PASS");
	my $login = getpwuid($<);
	my $uname = `uname -snrvm`;
	$uname =~ s/[\r\n]//g;
	my ($suffix) = get_journal_suffix_anchor($path);
	my $file_name = "$results_path/$results_dir/test_journals/lsb${type}chk.$suffix";
	if (open(JOURNAL_FILE, ">$file_name")) {
		print JOURNAL_FILE <<DATA;
0|3.7-lite $tm $date|User: $login, Command line: <none>
5|$uname|System Information
10|0 $path $tm|TC Start
15|0 tetj-1.0 1|TCM Start
400|0 1 1 $tm|IC Start
200|0 1 $tm|TP Start, Check file type
$result
410|0 1 1 $tm|IC End
80|0 0 $tm|TC End
900|$tm|TCC End
DATA
		close(JOURNAL_FILE);
	}
	else {
		print "    Error: Failed to create file $file_name\n    $!";
	}
}

sub create_journal_INFO($$$) {
	my ($path, $title_path, $type) = @_;
	my ($suffix, $anchor) = get_journal_suffix_anchor($path);
	my $file_name = "$results_path/$results_dir/test_journals/lsb${type}chk.$suffix.INFO";
	if (!-f $file_name) {
		if (open(INFO_FILE, ">$file_name")) {
			print INFO_FILE "TITLE: $title_path\nANCHOR: lsb${type}chk_$anchor\nJOURNAL_FILENAME: lsb${type}chk.$suffix\n";
			close(INFO_FILE);
			print "PROGRESS: Add: JOURNAL.\n" if ($verbose);
		}
		else {
			print "    Error: Failed to create file $file_name\n    $!";
		}
	}
	return "lsb${type}chk.$suffix";
}

# Looks for the unique temp directory name, creates it and returns its path
my $idx = 0;
sub create_temp_dir(;$) {
	my $prefix = ($_[0] or 'appcheck-temp');
	while (-e "/tmp/$prefix.$$-$idx") {
		++$idx;
	}
	mkdir("/tmp/$prefix.$$-$idx");
	return "/tmp/$prefix.$$-$idx";
}

# file_path => journal_name;
my %journals = ();

# File lists for testing with specific test programs
my @list_appchk = ();       # lsbappchk
my @list_appchk_sh = ();    # lsbappchk-shell
my @list_appchk_pl = ();    # lsbappchk-perl
my @list_appchk_py = ();    # lsbappchk-python
my @list_pkgchk = ();       # lsbpkgchk

my @list_db = ();           # File list to be passed to dependency analyser
my @clean_dirs = ();        # List of temp directories to be cleaned after execution

my @failed_files = ();      # List of files that caused problems

# Adds directory contents to the list (recursively)
sub add_dir_to_list($$);    # Prototype for recursive function
sub add_dir_to_list($$) {
	my ($dir, $file_list) = @_;
	if (opendir(DIR, $dir)) {
		my @elems = sort(grep(!/^\.+$/, readdir(DIR)));
		closedir(DIR);
		foreach (@elems) {
			if (-f "$dir/$_") {
				$file_list->{"$dir/$_"} = 1;
			}
			else {
				add_dir_to_list("$dir/$_", $file_list);
			}
		}
	}
}

# Reads the directory contents and process every file from it (recursively)
sub process_dir($$) {
	my ($path, $title_path) = @_;
	my %file_list = ();
	add_dir_to_list($path, \%file_list);
	foreach my $file (sort keys %file_list) {
		my $title = $file;
		$title =~ s/^\Q$path\E/$title_path|/;
		process_file($file, $title, 0);
	}
}

# Creates INFO file for the future journal and determines the test program for the file
# Input parameters:
# $path       - Real path to the file under test.
# $title_path - Visible path (including package from which the file or its parent package was unpacked).
# $root       - 1 if the file/package was specified explicitly by the user; 0 otherwise.
sub process_file($$$) {
	my ($path, $title_path, $root) = @_;
	my $temp_dir;
	my $path_esc = escape_path($path);
	my $type = detect_file_type($path);
	if ($type eq 'dir') {
		process_dir($path, $title_path);
	}
	elsif ($type eq 'tgz') {
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system("echo \"Unpacking TGZ package $path\" >>$tests_output");
		system("cd $temp_dir && tar xzf $path_esc 2>>$tests_output");
		if (!$?) {
			process_dir($temp_dir, $title_path);
		}
		else {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 1, 'app', 'Failed to unpack the archive.');
			}
		}
	}
	elsif ($type eq 'tbz') {
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system("echo \"Unpacking TBZ package $path\" >>$tests_output");
		system("cd $temp_dir && tar xjf $path_esc 2>>$tests_output");
		if (!$?) {
			process_dir($temp_dir, $title_path);
			$journals{$path} = create_journal_INFO($path, "$title_path|", 'app');
			create_journal($path, 101, 'app', 'TAR.BZ2 archives are not standardized by the LSB.');
		}
		else {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 1, 'app', 'Failed to unpack the archive.');
			}
		}
	}
	elsif ($type eq 'tar') {
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system("echo \"Unpacking TAR package $path\" >>$tests_output");
		system("cd $temp_dir && tar xf $path_esc 2>>$tests_output");
		if (!$?) {
			process_dir($temp_dir, $title_path);
		}
		else {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 1, 'app', 'Failed to unpack the archive.');
			}
		}
	}
	elsif ($type eq 'rpm') {
		$journals{$path} = create_journal_INFO($path, "$title_path|", 'pkg');
		push @list_pkgchk, $path;
		print "PROGRESS: Add: RPM.\n" if ($verbose);
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system("echo \"Unpacking RPM package $path\" >>$tests_output");
		system("cd $temp_dir && rpm2cpio $path_esc 2>>$tests_output | cpio -id 2>>$tests_output");
		if (!$?) {
			system("cd $utils_path_esc && ./pkg-scripts.pl -u RPM -d $temp_dir/SCRIPTS $path_esc 2>>$tests_output");
			process_dir($temp_dir, $title_path);
		}
		else {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 1, 'app', 'Failed to unpack the archive.');
			}
		}
	}
	elsif ($type eq 'deb') {
		$temp_dir = create_temp_dir('pkg-unpack');
		push @clean_dirs, $temp_dir;
		system("echo \"Unpacking DEB package $path\" >>$tests_output");
		system("cd $temp_dir && ar -x $path_esc 2>>$tests_output");
		if ($?) {
			if ($root) {
				push @failed_files, $title_path;
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 1, 'app', 'Failed to unpack the archive.');
			}
		}
		else {
			if ((-f "$temp_dir/data.tar.gz") or (-f "$temp_dir/data.tar.bz2")) {
				$temp_dir2 = create_temp_dir('pkg-unpack');
				push @clean_dirs, $temp_dir2;
				if (-f "$temp_dir/data.tar.gz") {
					system("cd $temp_dir2 && tar xzf $temp_dir/data.tar.gz 2>>$tests_output");
				}
				else {
					system("cd $temp_dir2 && tar xjf $temp_dir/data.tar.bz2 2>>$tests_output");
				}
				if (!$?) {
					system("cd $utils_path_esc && ./pkg-scripts.pl -u deb -d $temp_dir2/SCRIPTS $path_esc 2>>$tests_output");
					process_dir($temp_dir2, $title_path);
					$journals{$path} = create_journal_INFO($path, "$title_path|", 'app');
					create_journal($path, 101, 'app', 'DEB archives are not standardized by the LSB.');
				}
				else {
					if ($root) {
						push @failed_files, $title_path;
					}
					else {
						$journals{$path} = create_journal_INFO($path, $title_path, 'app');
						create_journal($path, 1, 'app', 'Failed to unpack the archive.');
					}
				}
			}
			else {
				$journals{$path} = create_journal_INFO($path, $title_path, 'app');
				create_journal($path, 101, 'app', 'No contents found in the DEB archive.');
			}
		}
	}
	elsif ($type eq 'elf-so') {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		push @list_appchk, $path;
		push @list_db, "$path|$title_path";
		print "PROGRESS: Add: ELF-so1.\n" if ($verbose);
		print "PROGRESS: Add: ELF-so2.\n" if ($verbose);
		print "PROGRESS: Add: APPDATA.\n" if ($verbose);
	}
	elsif ($type eq 'elf-bin') {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		push @list_appchk, $path;
		push @list_db, "$path|$title_path";
		print "PROGRESS: Add: ELF-bin.\n" if ($verbose);
		print "PROGRESS: Add: APPDATA.\n" if ($verbose);
	}
	elsif ($type eq 'elf-unsupported') {
		if ($root) {
			# The file was specified explicitly, we can't ignore it.
			# But it is unsupported by lsbappchk, so, report failure.
			$journals{$path} = create_journal_INFO($path, $title_path, 'app');
			create_journal($path, 1, 'app', 'Unsupported ELF type.');
			print "PROGRESS: Add-processed: ELF.\n" if ($verbose);
		}
		# Else just skip the file.
	}
	elsif ($type eq 'perl') {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		push @list_appchk_pl, $path;
		push @list_db, "$path|$title_path";
		print "PROGRESS: Add: Perl.\n" if ($verbose);
		print "PROGRESS: Add: APPDATA.\n" if ($verbose);
	}
	elsif ($type eq 'python') {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		push @list_appchk_py, $path;
		push @list_db, "$path|$title_path";
		print "PROGRESS: Add: Python.\n" if ($verbose);
		print "PROGRESS: Add: APPDATA.\n" if ($verbose);
	}
	elsif ($type eq 'shell') {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		push @list_appchk_sh, $path;
		push @list_db, "$path|$title_path";
		print "PROGRESS: Add: Shell.\n" if ($verbose);
		print "PROGRESS: Add: APPDATA.\n" if ($verbose);
	}
	elsif ($type =~ m/script:(.*)/) {
		$journals{$path} = create_journal_INFO($path, $title_path, 'app');
		create_journal($path, 1, 'app', 'Invalid script interpreter is used'.($1 ? " ($1)" : '').'.');
		print "PROGRESS: Add-processed: Shell.\n" if ($verbose);
	}
	elsif ($type =~ m/error:(.*)/) {
		if ($root) {
			push @failed_files, $title_path;
		}
		else {
			$journals{$path} = create_journal_INFO($path, $title_path, 'app');
			create_journal($path, 101, 'app', "Failed to read file contents.\n$1");
		}
	}
	elsif ($type eq 'unknown') {
		if ($root) {
			# The file was specified explicitly, we can't ignore it.
			# But we don't know how to test it, so, report failure.
			$journals{$path} = create_journal_INFO($path, $title_path, 'app');
			create_journal($path, 1, 'app', 'Failed to determine the file type.');
		}
		# Else just skip the file.
	}
	else {
		print "    Internal error: Bad return value from detect_file_type().\n    IN: '$path'\n    OUT: '$type'\n    Please report to lsb-appcheck-support\@linuxfoundation.org.\n";	}
}

# Create results directory
system("mkdir -p $results_path_esc/$results_dir_esc/test_journals");
system("mkdir -p $results_path_esc/$results_dir_esc/test_logs");
if ((!-d "$results_path/$results_dir/test_journals") or (!-d "$results_path/$results_dir/test_logs")) {
	$error = "Could not create subdirectories for storing the test results in\n$results_dir\n$!";
}
else {
	print "PROGRESS: Creating file list.\n" if ($verbose);
	# Create INFO files and fill file lists for testing
	foreach my $path (@paths) {
		process_file($path, $path, 1);
	}
	foreach my $pkg (@packages) {
		my @pkg_list;
		my $err_file = tmpnam();
		if (lc($package_manager) eq 'rpm') {
			system("rpm -q $pkg >$err_file 2>&1");
			if (!$?) {
				@pkg_list = `rpm -ql $pkg`;
			}
			else {
				create_journal_INFO("PKG: $pkg", "PKG: $pkg|", 'pkg');
				create_journal("PKG: $pkg", 1, 'pkg', "Could not find package:\n".`cat $err_file`);
			}
		}
		else {
			my $delim_pos = index($pkg, '_');
			my $pkg_name = (($delim_pos == -1) ? $pkg : substr($pkg, 0, $delim_pos));
			system("dpkg -l $pkg_name >$err_file 2>&1");
			if (!$?) {
				@pkg_list = `dpkg -L $pkg_name`;
			}
			else {
				create_journal_INFO("PKG: $pkg", "PKG: $pkg|", 'pkg');
				create_journal("PKG: $pkg", 1, 'pkg', "Could not find package:\n".`cat $err_file`);
			}
		}
		unlink($err_file);
		foreach (@pkg_list) {
			s/[\r\n]//g;
			process_file($_, "PKG: $pkg|$_", 0) if (!-d);
		}
	}
	my ($comp_list_fh, $comp_list_name) = tempfile("/tmp/$app_name.list.XXXXXX");
	my $comp_list_name_esc = escape_path($comp_list_name);
	if ($comp_list_fh) {
		foreach (@list_appchk) {
			print $comp_list_fh "$_\n";
		}
		close($comp_list_fh);
	}
	my ($comp_list_sh_fh, $comp_list_sh_name) = tempfile("/tmp/$app_name.list_sh.XXXXXX");
	my $comp_list_sh_name_esc = escape_path($comp_list_sh_name);
	if ($comp_list_sh_fh) {
		foreach my $file (@list_appchk_sh) {
			my $journal = $journals{$file};
			print $comp_list_sh_fh "SH: $file\t$results_path/$results_dir/test_journals/$journal\n";
		}
		close($comp_list_sh_fh);
	}
	my ($comp_list_pl_fh, $comp_list_pl_name) = tempfile("/tmp/$app_name.list_pl.XXXXXX");
	my $comp_list_pl_name_esc = escape_path($comp_list_pl_name);
	if ($comp_list_pl_fh) {
		foreach my $file (@list_appchk_pl) {
			my $journal = $journals{$file};
			print $comp_list_pl_fh "$file\t$results_path/$results_dir/test_journals/$journal\n";
		}
		close($comp_list_pl_fh);
	}
	my ($comp_list_db_fh, $comp_list_db_name) = tempfile("/tmp/$app_name.list_db.XXXXXX");
	my $comp_list_db_name_esc = escape_path($comp_list_db_name);
	if ($comp_list_db_fh) {
		foreach (@list_db) {
			print $comp_list_db_fh "$_\n";
		}
		close($comp_list_db_fh);
	}

	# Run tests
	if (scalar(@failed_files) == 0) {
		my $res;

		# lsbpkgchk
		print "PROGRESS: Starting RPM.\n" if ($verbose);
		foreach my $file (@list_pkgchk) {
			my $filename = $file;
			$filename =~ s!^.*/([^/]+)$!$1!;
			print "PROGRESS: Process: RPM ($filename).\n" if ($verbose);
			my $journal = escape_path($journals{$file});
			$file =~ s/[\r\n]//g;
			$file = escape_path($file);
			system("$tests_path_esc/lsbpkgchk -r $lsb_version -j $results_path_esc/$results_dir_esc/test_journals/$journal -T$lsb_profile $file >>$tests_output 2>&1");
		}

		# lsbappchk-shell
		print "PROGRESS: Starting Shell.\n" if ($verbose);
		if (scalar(@list_appchk_sh) > 0) {
			$res = 0;
			if (open(APPCHK_SH, "$tests_path_esc/lsbappchk-sh.pl --lsb=$lsb_version --list $comp_list_sh_name_esc 2>&1 |")) {
				my $line;
				my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
				while ($line = <APPCHK_SH>) {
					print TESTS_OUTPUT $line if ($out_opened);
					$line =~ s/[\r\n]//g;
					if ($verbose) {
						if ($line =~ m/^TESTING: (?:.*\/)?([^\/]+)$/) {
							print "PROGRESS: Process: Shell ($1).\n";
						}
					}
				}
				close(TESTS_OUTPUT) if ($out_opened);
				close(APPCHK_SH);
				$res = $?;
			}
			else {
				system("echo \"Failed to start lsbappchk-sh: $!\" >>$tests_output");
			}
			if ($res != 0) {
			    if ($res & 127) {
					system("echo \"Error executing lsbappchk-sh (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>$tests_output");
			    }
			    else {
					system("echo \"Unknown error: lsbappchk-sh exited with value ".($res >> 8)."!\" >>$tests_output");
			    }
			}
		}

		# lsbappchk-perl
		print "PROGRESS: Starting Perl.\n" if ($verbose);
		if (scalar(@list_appchk_pl) > 0) {
			$res = 0;
			if (open(APPCHK_PL, "$tests_path_esc/lsbappchk.pl -v $lsb_version -j --list $comp_list_pl_name_esc --verbose 2>&1 |")) {
				my $line;
				my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
				while ($line = <APPCHK_PL>) {
					print TESTS_OUTPUT $line if ($out_opened);
					$line =~ s/[\r\n]//g;
					if ($verbose) {
						if ($line =~ m/^TESTING: (?:.*\/)?([^\/]+)$/) {
							print "PROGRESS: Process: Perl ($1).\n";
						}
					}
				}
				close(TESTS_OUTPUT) if ($out_opened);
				close(APPCHK_PL);
				$res = $?;
			}
			else {
				system("echo \"Failed to start lsbappchk-perl: $!\" >>$tests_output");
			}
			if ($res != 0) {
			    if ($res & 127) {
					system("echo \"Error executing lsbappchk-perl (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>$tests_output");
			    }
			    else {
					system("echo \"Unknown error: lsbappchk-perl exited with value ".($res >> 8)."!\" >>$tests_output");
			    }
			}
		}

		my $temp_dir = create_temp_dir('appcheck-journals');
		my $old_dir = cwd();
		chdir($temp_dir);

		# lsbappchk-python
		print "PROGRESS: Starting Python.\n" if ($verbose);
		foreach my $file (@list_appchk_py) {
			my $filename = $file;
			$filename =~ s!^.*/([^/]+)$!$1!;
			print "PROGRESS: Process: Python ($filename).\n" if ($verbose);
			my $journal = escape_path($journals{$file});
			$file =~ s/[\r\n]//g;
			$file = escape_path($file);
			system("$tests_path_esc/lsbappchk.py -v $lsb_version -j $file >>$tests_output 2>&1");
			my @journals_res = `ls`;
			if (scalar(@journals_res) > 1) {
				print "    Warning: More than one journal were generated by lsbappchk-python!\n    Only first one will be used.\n";
			}
			$journals_res[0] =~ s/[\r\n]//g;
			$journals_res[0] = escape_path($journals_res[0]);
			system("mv $journals_res[0] $results_path_esc/$results_dir_esc/test_journals/$journal >>$tests_output 2>&1");
			system('rm -rf ./*');
		}
		chdir($old_dir);
		rmdir($temp_dir);

		# lsbappchk
		$res = 0;
		if (open(APPCHK_ELF, "$tests_path_esc/lsbappchk -r $lsb_version -j -o $results_path_esc/$results_dir_esc/test_journals -l $comp_list_name_esc -T$lsb_profile 2>&1 |")) {
			my $line;
			my %stages = ('SO-Pass1' => 'ELF-so1', 'SO-Pass2' => 'ELF-so2', 'BIN' => 'ELF-bin');
			my $stage = '';
			my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
			while ($line = <APPCHK_ELF>) {
				print TESTS_OUTPUT $line if ($out_opened);
				$line =~ s/[\r\n]//g;
				if ($verbose) {
					if ($line =~ m/^(SO-Pass1|SO-Pass2|BIN): .*\/([^\/]+)$/) {
						if ($stage ne $stages{$1}) {
							$stage = $stages{$1};
							print "PROGRESS: Starting $stage.\n";
						}
						print "PROGRESS: Process: $stage ($2).\n";
					}
				}
			}
			close(TESTS_OUTPUT) if ($out_opened);
			close(APPCHK_ELF);
			$res = $?;
		}
		else {
			system("echo \"Failed to start lsbappchk: $!\" >>$tests_output");
		}
		if ($res != 0) {
		    if ($res & 127) {
				system("echo \"Error executing lsbappchk (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>$tests_output");
		    }
		    else {
				system("echo \"Unknown error: lsbappchk exited with value ".($res >> 8)."!\" >>$tests_output");
		    }
		}

		print "PROGRESS: Starting JOURNAL.\n" if ($verbose);

		# Save information about test run
		system("touch $results_path_esc/HISTORY");
		if (open(FILE, "+<$results_path/HISTORY")) {
			flock(FILE, LOCK_EX);
			seek(FILE, 0, SEEK_END);
			print FILE "$results_dir\n";
			flock(FILE, LOCK_UN);
			close(FILE);
		}
		else {
			system("echo \"Could not open HISTORY file!\nThe test results will not be listed on the Results page.\" >>$tests_output 2>&1");
		}
		if (open(COMMENTS_FILE, ">$results_path/$results_dir/test_logs/test_comments")) {
			print COMMENTS_FILE $comments;
			close(COMMENTS_FILE);
		}
		system("echo \"$lsb_version\" >> $results_path_esc/$results_dir_esc/test_logs/lsb_version");
		system("echo \"$architecture\" >> $results_path_esc/$results_dir_esc/test_logs/host_platform");

		# Generate the report
		if (-s "$results_path/$results_dir/test_logs/tests_output" != 0) {
			system("echo \"\" >>$tests_output");
		}
		$res = 0;
		if (open(GEN_REPORT, "$utils_path_esc/report.pl $results_path_esc/$results_dir_esc -n $app_name_esc -o $results_dir_esc.tgz 2>&1 |")) {
			my $line;
			my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
			while ($line = <GEN_REPORT>) {
				print TESTS_OUTPUT $line if ($out_opened);
				$line =~ s/[\r\n]//g;
				if ($line =~ m/^Reading journal: test_journals\/(.*)$/) {
					print "PROGRESS: Process: JOURNAL ($1).\n" if ($verbose);
				}
			}
			close(TESTS_OUTPUT) if ($out_opened);
			close(GEN_REPORT);
			$res = $?;
		}
		else {
			system("echo \"Failed to start report.pl: $!\" >>$tests_output");
		}
		if ($res != 0) {
		    if ($res & 127) {
				system("echo \"Error executing report.pl (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>$tests_output");
		    }
		    else {
				system("echo \"Unknown error: report.pl exited with value ".($res >> 8)."!\" >>$tests_output");
		    }
		}

		# Collects application data
		# apptodb needs PM file located in its directory.
		print "PROGRESS: Starting APPDATA.\n" if ($verbose);
		chdir($utils_path);
		system("echo \"\" >>$tests_output");
		system("echo \"Collecting data about application files\" >>$tests_output");
		$res = 0;
		if (open(APPTODB, "$utils_path_esc/apptodb.pl --verbose --trsp-symlink -a $architecture -l $comp_list_db_name_esc -o $results_path_esc/$results_dir_esc/${app_name_esc}_${f_architecture}_upload_data $app_name_esc 2>&1 |")) {
			my $line;
			my $out_opened = open(TESTS_OUTPUT, ">>$tests_output");
			while ($line = <APPTODB>) {
				print TESTS_OUTPUT $line if ($out_opened);
				$line =~ s/[\r\n]//g;
				if ($line =~ m/^Processing file: '.*\/([^\/]+)'$/) {
					print "PROGRESS: Process: APPDATA ($1).\n" if ($verbose);
				}
			}
			close(TESTS_OUTPUT) if ($out_opened);
			close(APPTODB);
			$res = $?;
		}
		else {
			system("echo \"Failed to start apptodb.pl: $!\" >>$tests_output");
		}
		if ($res != 0) {
		    if ($res & 127) {
				system("echo \"Error executing apptodb.pl (terminated with signal ".($res & 127).(($res & 128) ? ", with coredump" : "").")!\" >>$tests_output");
		    }
		    else {
				system("echo \"Unknown error: apptodb.pl exited with value ".($res >> 8)."!\" >>$tests_output");
		    }
		}

		# Generates application analysis files
		print "PROGRESS: Starting MERGE.\n" if ($verbose);

		my $updata_file = $results_path_esc."/".$results_dir_esc."/".$app_name_esc."_".$f_architecture."_upload_data";

		# Analysis
		system("$utils_path_esc/analysis.pl $lsb_version $architecture $updata_file >>$tests_output 2>&1");

		print "PROGRESS: Starting CLEANUP.\n" if ($verbose);
	}
	else {
		$error = "Could not process the following files:\n".join("\n", @failed_files);
	}
	unlink($comp_list_name);
	unlink($comp_list_sh_name);
	unlink($comp_list_pl_name);
	unlink($comp_list_db_name);
	foreach (@clean_dirs) {
		system("rm -rf $_ >/dev/null 2>&1");
	}
	print "PROGRESS: Starting FINISH.\n" if ($verbose);
}

if (defined($error)) {
	system("rm -rf $results_path_esc/$results_dir_esc >/dev/null 2>&1");
	print "ERROR: ".join("\nERROR: ", split(/\n/, $error))."\n";
	exit 4;
}

print "RESULT: $results_dir\n";
exit 0;
